# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath('..'))


# -- Project information -----------------------------------------------------

project = 'astro-sedpy'
copyright = '2014-2022, Benjamin Johnson and Contributors'
author = 'Benjamin Johnson'

# The full version, including alpha/beta/rc tags
release = '0.3'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.mathjax',
    "sphinx.ext.intersphinx",
    "myst_nb",
    'sphinx.ext.napoleon',
    'numpydoc'
]

myst_enable_extensions = ["dollarmath", "colon_fence"]

# Add any paths that contain templates here, relative to this directory.
#templates_path = ['_templates']

language = None
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']
autodoc_mock_imports = []
# The name of the Pygments (syntax highlighting) style to use.
#pygments_style = 'sphinx'
# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_title = "sedpy"
html_theme = 'sphinx_book_theme'
html_copy_source = True
html_show_sourcelink = True
html_theme_options = {"path_to_docs": "docs",
                      "repository_url": "https://github.com/bd-j/sedpy",
                      "repository_branch": "main",
                      "use_repository_button": True,
                      "use_edit_page_button": True,
                      "use_issues_button": True,
                      "use_download_button": True}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
#html_static_path = ['_static']
#html_css_files = ['css/custom.css']
#html_logo = "_static/logo_name.png"
#html_favicon = "_static/favicon.png"

# If not None, a 'Last updated on:' timestamp is inserted at every page
# bottom, using the given strftime format.
# The empty string is equivalent to '%b %d, %Y'.
html_last_updated_fmt = ''



# -*- coding: utf-8 -*-

"""photometr.py - Simple Aperture photometry.  Very old code, superceded by
astropy affiliated package `photutils.`
"""

# FIXME: kind of a stupid class dependence.
# Ideally a photometer object should take an image and a region object
# as arguments, where the region object is an instance of a particualr aperture class and
# can return an in_region boolean (or perhaps a 'fraction') for
# any position(s).  As it is now the 'photometer' object (called Aperture)
# is actually subclassed by the more detailed apertures/regions,
# and requires passing a shape dictionary as well.  redundant

import sys
import numpy as np
from numpy import hypot, sqrt

#the below are for some of the more arcane sky measurement methods
try:
    from scipy.optimize import curve_fit
    import sklearn
    from astroML.density_estimation import bayesian_blocks
    import matplotlib.pyplot as pl
except ImportError:
    pass


thismod = sys.modules[__name__]


class Photometer(object):
    """
    Trying for a better class dependence.  Basically wraps the image
    in an object with photometry methods.  Incomplete
    """
    def __init__(self, image, wcs=None, ivar=None):
        self.nx, self.ny = image.shape
        self.image = image.flatten()
        self.wcs = wcs
        self.ivar = ivar
        yy, xx = np.indices(self.nx, self.ny)
        if wcs is not None:
            self._x, self._y = wcs.wcs_pix2world(xx, yy, 0)
        else:
            self._x, self._y = xx, yy

    def measure_flux(self, aperture, background):
        """
        Measure background subtracted flux.  Takes an aperture object,
        and a local background object.
        """
        o, a, e = self.object_flux(aperture)
        b, ba, be = background.evaluate(self)
        flux = o - a*b
        flux_var = e*e + a*a*be*be/ba
        return flux, sqrt(flux_var)

    def object_flux(self, aperture, weights=1.0):
        """
        Measure total flux (source + background) within an aperture.
        Takes an image, and an aperture object.
        """
        fracs = aperture.contains(self._x, self._y)
        inds = fracs > 0
        if self.ivar is not None:
            unc = sqrt((fracs[inds]/self.ivar[inds]).sum())
        else:
            unc = np.nan

        return (self.image * weights * fracs).sum(), fracs.sum(), unc


class Aperture(object):

    def world_to_pixels(self, shape, wcs):

        pass

    def object_flux(self, shape, image, ivar=None):
        """Measure total flux within an aperture (source + background)"""
        inds, fracs = self.pixnum(**shape)
        unc = 0
        if ivar is not None:
            unc = sqrt((fracs/ivar[inds[0], inds[1]]).sum())
        return (image[inds[0], inds[1]]*fracs).sum(), fracs.sum(), unc

    def measure_flux(self, shape, image, wcs=None, skypars=None, ivar=None):
        """Measure background subtracted flux."""
        o, a, e = self.object_flux(shape, image, ivar=ivar)
        b, ba, be = self.background.evaluate(image, skypars)
        flux = o - a*b
        flux_var = e*e + a*a*be*be/ba
        return flux, sqrt(flux_var)

    #def get_flux(self, image, ivar = None):
    #    return self.measure_flux(self.shape, image, ivar = ivar, skypars = self.skypars, wcs = self.wcs)


class Circular(Aperture):

    def __init__(self, exact=False):
        if exact is True:
            self.pixnum = circle_frac_exact
        else:
            self.pixnum = circle_frac_quick
        self.background = ZeroSky()


class Elliptical(Aperture):

    def __init__(self):
        self.pixnum = ellipse_frac_quick()
        self.background = ZeroSky()


class Box(Aperture):

    def __init__(self):
        self.pixnum = box_frac_quick()
        self.background = ZeroSky()


# --- Classes for sky measurement ---


class Background(object):

    def evaluate(self, image, skypars):
        inds, fracs = self.pixnum(**skypars)
        value, sdpp = self.skystats(image[inds[0], inds[1]], **skypars)
        return value, len(inds), sdpp


class Annulus(Background):

    def __init__(self, bgtype='quartile_sky'):
        self.pixnum = circle_frac_quick
        self.skystats = getattr(thismod, bgtype)


class EllipticalAnnulus(Background):

    def __init__(self, bgtype='quartile_sky'):
        self.pixnum = ellipse_frac_quick
        self.skystats = getattr(thismod, bgtype)


class ZeroSky(Background):
    """A class for sky values of zero, or for user defined sky statistics.
    The return_value is a tuple giving (sky, sky_area, sigma_sky_per_pixel)"""

    def __init__(self, bgtype='quartile_sky', return_value=(0, 1, 0)):
        self.pixnum = None
        self.skystats = None
        self.return_value = return_value

    def evaluate(self, image, skypars):
        return self.return_value


# --- Pixnum methods ---


def circle_frac_quick(xcen=0, ycen=0, radius=1, inner_radius=None, subpixels=1,
                      **extras):
    """obtain fractional pixel coverage.  optionally use subpixels to
    increase precison (though this doesn't seem to help).  Assumes pixel
    centers have coordinates X.5, Y.5 """
    #setup
    center = np.array([xcen, ycen])
    sz = np.ceil((radius+1)*2)
    start = np.floor(center + 0.5 - radius)
    center = center*subpixels
    radius = radius*subpixels
    sz = sz*subpixels
    start = (start-1)*subpixels
    if (start < 0).any():
        raise ValueError('Aperture extends off image edge')
    off = center - start - 0.5
    yy, xx = np.ogrid[0:sz, 0:sz]
    rr = hypot(xx - off[0], yy-off[1])

    #find pixels within the radius
    within = (radius+0.5) - rr
    within[within > 1.0] = 1.0
    within[within < 0] = 0.

    #if it's an annulus
    if inner_radius is not None:
        within_inner = inner_radius*subpixels + 0.5 - rr
        within_inner[within_inner < 0.0] = 0.0
        within_inner[within_inner > 1.0] = 1.0
        within = within - within_inner
    an = within

    #rebin if you used subpixels
    if subpixels != 1:
        an = an.reshape((an.shape[0]/subpixels, subpixels, an.shape[1]/subpixels, subpixels)).mean(1).mean(2)
    #pick the pixels to rturn, and get their fractional coverage
    pix1 = np.where(an > 0.0)
    fracs = an[pix1[0], pix1[1]]
    x = (pix1[0] + start[0]/subpixels).astype('i8')
    y = (pix1[1] + start[1]/subpixels).astype('i8')

    return (x, y), fracs


def circle_frac_exact(xcen, ycen, radius):
    pass


def ellipse_frac_quick(xcen=0, ycen=0, a=1, b=1, pa=0, precision=None):

    yy, xx = np.ogrid[0:sz, 0:sz]
    dx, dy = (xx - off[0]), (yy - off[1])

    within = 1 - np.sqrt(((dx * np.cos(pa) - dy * np.sin(pa))/a)**2 +
                         ((dx * np.sin(pa) + dy * np.cos(pa))/b)**2)
    within[within > 1.0] = 1.0
    within[within < 0] = 0.


    #rebin if you used subpixels
    if subpixels != 1:
        an = an.reshape((an.shape[0] / subpixels, subpixels,
                         an.shape[1] / subpixels, subpixels)).mean(1).mean(2)
    #pick the pixels to rturn, and get their fractional coverage
    pix1 = np.where(an > 0.0)
    fracs = an[pix1[0], pix1[1]]
    x = (pix1[0] + start[0] / subpixels).astype('i8')
    y = (pix1[1] + start[1] / subpixels).astype('i8')

    return (x, y), fracs


# --- SKY statistics determination methods ---


def quartile_sky(values, percentiles=[0.16, 0.5, 0.84], **extras):
    """Use the median and 16th percentile to estimate the standard
    deviation per pixel."""

    percentiles = np.asarray(percentiles)
    npix = len(values)
    #oo = np.argsort(values)
    qval = np.sort(values)[np.round(npix*percentiles).astype('i8')]
    #qval = values[oo[np.round(npix*percentiles)]]
    return qval[1], qval[1]-qval[0]


def gaussfit_sky(values, p_thresh=0.65, plot=False, **extras):
    """Fit a gaussian to the lower part of a histogram of the sky values.
    The histogram bins are estimated using Bayesian blocks.  p_thresh gives
    the percentile below which the gaussian is fitted to the data. Return
    central value and estimate of standard deviation per pixel """

    bins = bayesian_blocks(values)
    print(len(bins), bins)
    #dbin = bins[1:]-bins[:-1]
    cbin = (bins[1:]+bins[:-1])/2
    hist = np.histogram(values, bins=bins, range=(bins.min(), bins.max()), density=True)

    #pdf = hist/dbin
    val_thresh = np.percentile(values, p_thresh)
    lower = cbin < p_thresh

    def gauss(x, *p):
        A, mu, sigma = p
        return A*np.exp(-(x-mu)**2/(2.*sigma**2))

    # p0 is the initial guess for the fitting coefficients (A, mu and sigma above)
    p0 = [np.max(hist[0]), values.mean(), values.std()]
    coeff, var_matrix = curve_fit(gauss, cbin[lower], hist[0][lower], p0=p0)
    if plot:
        print(len(hist[1]), len(hist[0]), type(coeff))
        pl.figure()
        pl.plot(cbin, hist[0], color='b')
        pl.plot(cbin, gauss(cbin, [coeff[0], coeff[1], coeff[2]]), color='r')
        pl.axvline(val_thresh)
    return coeff[1], coeff[2]


def gmm_sky(values, **extras):
    """Use a gaussian mixture model, via expectation maximization.
    of course, there's only one gaussian.  could add another for
    faint sources, bad pixels, but..."""

    gmm = sklearn.mixture.GMM()
    r = gmm.fit(values)
    return r.means_[0, 0], np.sqrt(r.covars_[0, 0])


def sigclip_sky(values, sigma=[3, 2.25], minlength=5, **extras):
    """Use iterative sigma clipping"""

    def between(vals, sigs):
        m, s = vals.mean(), vals.std()
        return (vals < (m + sigs[1] * s)) & (vals > (m - sigs[0] * s))

    while ((False in between(values, sigma)) & (len(values) > minlength)):
        values = values[between(values, sigma)]

    return values.mean(), values.std()

# --- Centroiding ---


def centroid(images):
    """Dumb dumb centroiding.  assumes x and y axes are the
    last two dimensions of images.  Something is wrong with the
    broadcasting.  absolutely *have* to include weighting"""
    sz = images.shape[-2:]
    xg = np.arange(sz[0])
    yg = np.arange(sz[1])
    denom = images.sum(axis=(-1, -2))
    y = (yg[None, None, :] * images).sum(axis=(-2, -1)) / denom
    x = (xg[None, :, None] * images).sum(axis=(-2, -1)) / denom
    return x, y



# -*- coding: utf-8 -*-

"""modelgrid.py  - Deprecated code for interpolating regular or irregular grids
of models.  Use `scipy.interpolate.RegularGridInterpolator` or
`scipy.interpolate.LinearNDInterpolator`.

Kept here since it is used by some other (old) packages of mine.
"""

import numpy as np
from scipy.spatial import Delaunay
try:
    import astropy.io.fits as pyfits
except (ImportError):
    import pyfits
try:
    import sklearn.neighbors
except(ImportError):
    pass
try:
    from . import observate
except (ImportError):
    print('Warning - observate not imported, SpecLibrary class unavailable')


class ModelLibrary(object):
    """Class to deal with (irregular) grids of models.  Primary
    attribute is `pars`: a structured array of parameter values of
    shape (ngrid).  Methods are provided to manipulate the parameter
    structure, and to obtain interpolation weights for specific points
    within the grid using a variety of methods.
    """
    # force generation of the DT and KDtree in first pass
    triangle_dirtiness = 1

    def __init__(self, pars=None, parnames=None):
        if pars is not None:
            self.set_pars(pars, parnames)

    def set_pars(self, pars, parnames):
        self.pars = self.structure_array(pars, parnames)
        self.ngrid = self.pars.shape[0]

    def add_par(self, value, name, dtype='<f8'):
        newp = self.structure_array(value, [name], types=[dtype])
        self.pars = self.join_struct_arrays([self.pars, newp])
        pass

    def par_names(self):
        return self.pars.dtype.names

    def par_range(self, parname, inds=None):
        range_list = [(np.nanmin(self.pars[inds][p]),
                       np.nanmax(self.pars[inds][p]), p)
                      for p in parname]
        return range_list

    def structure_array(self, values, fieldnames, types=None):
        """Turn a numpy array of floats into a structurd array.

        :param values: ndarray, shape(nrec, nfield).
            Values of the structure.

        :param fieldnames: string sequence of length nfield.
            A list or string array of parameter names with length
            NFIELD.

        :param types: string sequence, optional
            Format identifiers for each field. Defaults to '<f8'.

        :returns struct:
            A numpy structured array.
        """
        # shape coercion
        values = np.atleast_2d(values)
        if values.shape[-1] != len(fieldnames):
            if values.shape[0] == len(fieldnames):
                values = values.T
            else:
                raise ValueError('array and fieldnames do not have consistent shapes!')
        nobj = values.shape[0]

        # create the dtype and structured array
        if types is None:
            types = ['<f8'] * len(fieldnames)
        dt = np.dtype(zip(fieldnames, types))
        struct = np.zeros(nobj, dtype=dt)
        for i, f in enumerate(fieldnames):
            struct[f] = values[..., i]
        return struct

    def join_struct_arrays(self, arrays):
        """From some dudes on StackOverflow.  Add equal length
        structured arrays to produce a single structure with fields
        from both.

        :param arrays:
            A sequence of structured arrays.  These must each have the
            same length.

        :returns newarray:
            A single array containing the fields of ``arrays``.
        """
        if False in [len(a) == len(arrays[0]) for a in arrays]:
            raise ValueError('array lengths do not match.')

        newdtype = np.dtype(sum((a.dtype.descr for a in arrays), []))
        if len(np.unique(newdtype.names)) != len(newdtype.names):
            raise ValueError('arrays have duplicate fields.')
        newrecarray = np.empty(len(arrays[0]), dtype=newdtype)
        for a in arrays:
            for name in a.dtype.names:
                newrecarray[name] = a[name]
        return newrecarray

    def model_weights(self, target_points, parnames=None, subinds=None,
                      itype='dt', force_triangulation=False, **extras):
        """Given an array of target coordinates and optionally a list
        of parnames, return the indices and weights of the model grid
        points that will interpolate to the target points.

        :param target_points: ndarray, shape(ntarg, npar)
            The points to which you want to interpolate.  Can be a
            numpy structured array of length NTARG or a simple numpy
            array of dimension [NTARG x NPAR], if parnames is also
            passed.  If a numpy structured array, then the field names
            should correspond to parameters in the `par` structure.
            1d arrays (i.e. a single target point) will be
            automatically upgraded to 2d arrays

        :param parnames: string sequence
            A list of the parameter names corresponding to the NPAR
            dimension of target_points.  These names should correspond
            to parameters in the `par` structure.

        :param subinds:
            An optional array identifying the members of the grid to
            be used for interpolation.

        :param itype: string (default 'dt')
            A string giving the type of interpolation to perform.
            Available options are:

            * 'dt' - use Delaunay triangulation.  This is suitable for
              irregular grids, though care must be taken that target
              points are within the convex hull described by the grid
            * 'idw' - inverse distance weighting to the NPAR nearest
              neighbors.
            * 'nearest' - nearest neighbor, as determined by a kd-tree
            * 'ndlinear' - not yet implemented'

        :param force_triangulation: bool (default True)
            Normally the triangulation and kdtree formed from the
            model grid is stored and reused unless the
            `triangle_dirtiness' attribute is greater than zero.  If
            this keyword is true, force the triangulation and kdtree
            to be regenerated regardless of 'dirtiness'

        :returns inds: ndarray, shape(ntarg, nind)
            An array that indexes the model parameter grid to give the
            interpolating models for each target point.  It has shape
            (NTARG, NIND) where nind is NPAR+1 for Delaunay
            triangulation, NPAR for 'idw', and 1 for 'nearest'

        :returns weights: ndarray, shape(ntarg, nind)
            The interpolation weight for each model grid point
            specified by inds, and of identical shape. The weights
            summed along the NIND direction will always be 1. Thus for
            'nearest' weights is always 1.
        """
        # deal with recarray input
        if parnames is None:
            parnames = target_points.dtype.names
            targets = np.array(target_points.tolist())
        else:
            targets = target_points
        targets = np.atleast_2d(targets)

        # If necessary rebuild the model delauynay triangulation and
        # KDTree.  Pull the grid points out of the model record data
        # and make an (nmodel,ndim) array of model grid parameter
        # values.  Need to loop to make sure order is correct.
        # if subinds is not None:
        #     force_triangulation = True
        # print( subinds)
        self.triangle_dirtiness += force_triangulation
        if self.triangle_dirtiness > 0:
            self.refresh_graphs(parnames, subinds=subinds)

        # pass the result to weightsDT
        if itype.lower() == 'dt':
            inds, weights = self.weightsDT(targets)
        elif itype.lower() == 'idw':
            inds, weights = self.weights_kNN_inverse_dist(targets, k=targets.shape[1])
        elif itype.lower() == 'nearest':
            inds, weights = self.weights_kNN_inverse_dist(targets, k=1)
        return inds, weights

    def refresh_graphs(self, parnames, subinds=None):
        """Given parameter names and optionally specific model
        indices, build a a Delaunay triangulation and a KDTree for the
        model points.
        """
        model_points = [np.squeeze(self.pars[subinds][pname])
                        for pname in parnames]
        model_points = np.array(model_points).transpose()  # (nmod, ndim)
        self.graphed_parameters = parnames
        self.triangle_dirtiness = 0
        # Delaunay triangulate
        self.dtri = Delaunay(model_points)
        # kdtree
        try:
            self.kdt = sklearn.neighbors.KDTree(model_points)
        except(NameError):
            pass

    def weightsDT(self, target_points):
        """The interpolation weights are determined from barycenter
        coordinates of the vertices of the enclosing Delaunay
        triangulation simplex. This allows for the use of irregular Nd
        grids. See also weights_1DLinear and
        weights_kNN_inverse_distance.

        :param target_points: ndarray, shape(ntarg,npar)
            The coordinates to which you wish to interpolate.

        :returns inds: ndarray, shape(ntarg,npar+1)
             The model indices of the interpolates.

        :returns weights: narray, shape (ntarg,npar+1)
             The weights of each model given by ind in the
             interpolates.
        """
        # Find the encompassing (hyper)triangle(s) for the desired
        # points, given a delauynay triangulation
        ndim = target_points.shape[-1]
        # Output triangle_inds is an (ntarg) array of simplex indices
        triangle_inds = self.dtri.find_simplex(target_points)
        # And get model indices of (hyper)triangle vertices. inds has
        # shape (ntarg,ndim+1)
        inds = self.dtri.vertices[triangle_inds, :]
        # get the barycenter coordinates through matrix multiplication
        # and dimensional juggling
        tmp = self.dtri.transform[triangle_inds, ndim, :]
        bary = np.dot(self.dtri.transform[triangle_inds, :ndim, :ndim],
                      (target_points - tmp).reshape(-1, ndim, 1))
        oned = np.arange(triangle_inds.shape[0])
        # ok.  in np 1.7 can add an axis to squeeze
        bary = np.atleast_2d(np.squeeze(bary[oned, :, oned, :]))
        # the last bary coordinate is 1-sum of the other coordinates
        last = 1 - bary.sum(axis=-1)
        weights = np.hstack((bary, last[:, np.newaxis]))

        outside = (triangle_inds == -1)
        weights[outside, :] = 0

        return inds, weights

        # loop implementation of the above for clarity
        #   npts = triangle_inds.shape[0]
        #   bary = np.zeros([npts,ndim+1])
        #   for i in xrange(npts):
        #      bary[i,:-1]= np.dot( dtri.transform[triangle_inds[0],:ndim,:ndim],
        #                         (target_points-dtri.transform[triangle_inds[i],ndim,:])

    def weights_kNN_inverse_dist(self, target_points, k=1):
        """The interpolation weights are determined from the inverse
        distance to the k nearest neighbors.

        :param target_points: ndarray, shape(ntarg,npar)
            The coordinates to which you wish to interpolate.

        :param k:
            The number of nearest neighbors to use.

        :returns inds: ndarray, shape(ntarg,npar+1)
             The model indices of the interpolates.

        :returns weights: narray, shape (ntarg,npar+1)
             The weights of each model given by ind in the
             interpolates.
        """
        dists, inds = self.kdt.query(target_points, k=k,
                                     return_distance=True)
        if k == 1:
            return inds, np.ones(inds.shape)
        weights = 1 / dists
        weights[np.isinf(weights)] = large_number
        weights = weights/weights.sum(axis=-1)
        return inds, weights

    def weights_1DLinear(self, model_points, target_points,
                         extrapolate=False, left=0.0, right=0.0):
        """The interpolation weights are determined from 1D linear
        interpolation.

        :param model_points: ndarray, shape(nmod)
            The parameter coordinate of the available models

        :param target_points: ndarray, shape(ntarg)
            The coordinate to which you wish to interpolate

        :returns inds: ndarray, shape(ntarg,2)
             The model indices of the interpolates

        :returns weights: narray, shape (ntarg,2)
             The weights of each model given by ind in the interpolates.
        """
        order = model_points.argsort()
        mod_sorted = model_points[order]

        x_new_indices = np.searchsorted(mod_sorted, target_points)
        x_new_indices = x_new_indices.clip(1, len(mod_sorted) - 1).astype(int)
        lo = x_new_indices - 1
        hi = x_new_indices
        x_lo = mod_sorted[lo]
        x_hi = mod_sorted[hi]
        width = x_hi - x_lo
        w_lo = (x_hi - target_points) / width
        w_hi = (target_points - x_lo) / width

        if extrapolate is False:
            # find places where target is above or below the model
            # range.
            above_scale = w_lo < 0
            below_scale = w_hi < 0
            # set the indices to be indentical in these cases
            lo[above_scale] = hi[above_scale]
            hi[below_scale] = lo[below_scale]
            # make the combined weights sum to ``left`` or ``right``
            w_lo[above_scale] = 0
            w_hi[above_scale] = left
            w_hi[below_scale] = 0
            w_lo[below_scale] = right

        inds = np.vstack([lo, hi]).T
        weights = np.vstack([w_lo, w_hi]).T

        return inds, weights

    def nearest_index(self, array, value):
        return (np.abs(array-value)).argmin(axis=-1)


class SpecLibrary(ModelLibrary):
    """Class to operate on spectral libraries. Methods are provided to
    interpolate the available model spectra (stored as a structured
    parameter array and a spectral array) to a certain set of
    parameters.  Subclasses are used to return the actual model
    spectrum given a set of model parameters.  Primary attributes are
    pars, spectra, wavelength.  Spectra should be of shape (NOBJ x
    NWAVE)
    """
    flux_unit = 'erg/s/cm^2/AA of 1solar mass at 10pc'

    def __init__(self):
        pass

    def spectra_from_pars(self):
        """This should take a numpy structured array of parameters of
        length nobj and return an (nobj, nwave) array of spectra
        """
        raise(NotImplementedError)

    def generateSEDs(self, pars, filterlist, wave_min=90, wave_max=1e7,
                     keepspec=False, intspec=False, attenuator=None, **extras):
        """
        :returns sed:
            ndarray of shape (nobj,nfilter)

        :returns lbol:
            ndarray of shape (nobj)

        :returns outspectra:
            ndarray of shape (nobj,nwave)
        """
        # don't use too much memory at once
        maxmod = 1e7/self.wavelength.shape[0]
        ngrid = pars.shape[0]
        sed = np.zeros([int(ngrid), len(filterlist)])
        lbol = np.zeros(int(ngrid))
        outspectra = None
        if keepspec:
            outspectra = np.zeros([ngrid, self.wavelength.shape[0]])
        elif intspec:
            outspectra = np.zeros(self.wavelength.shape[0])

        # split big model grids to avoid memory constraints
        i = 0
        while (i*maxmod <= ngrid):
            s1, s2 = int((i)*maxmod), int(np.min([(i+1)*maxmod-1, ngrid]))
            spec = self.spectra_from_pars(pars[int(s1):int(s2)], **extras)
            if attenuator is not None:
                spec = attenuator.attenuate_spectrum(self.wavelength, spec,
                                                     pars[s1:s2], **extras)
            sed[s1:s2, :] = observate.getSED(self.wavelength, spec, filterlist)
            lbol[s1:s2] = observate.Lbol(self.wavelength, spec, wave_min=wave_min,
                                         wave_max=wave_max)
            i += 1
            if keepspec:
                outspectra[s1:s2, :] = spec
            elif intspec:
                outspectra += spec.sum(axis=0)

        return sed, lbol, outspectra

    def interpolate_to_pars(self, target_points, parnames=None,
                            subinds=None, itype='dt', **extras):
        """Method to obtain the model spectrum for a given set of
        parameter values via interpolation of the model grid. The
        interpolation weights are determined from barycenters of a
        Delaunay triangulation or nLinear interpolation or
        k-nearest-neighbor inverse distance.

        :param target_points: ndarray, shape (ntarg,ndim)
            Desired model parameters. Can also be a structured array
            with fields named for the model parameters.  See
            ModelLibrary.model_weights()

        :param subinds: ndarray
            Indices of the model pars structure to use in
            interpolation. Allows for only portions of the model
            library to be used.

        :param parnames: string sequence, len (ndim)
            The names of the model library parameters

        :returns spectra: ndarray, shape (ntarg, nwave)
            The interpolated spectra.
        """
        inds, weights = self.model_weights(target_points, parnames=parnames,
                                           itype=itype, subinds=subinds, **extras)
        if subinds is not None:
            inds = subinds[inds]
        return self.combine_weighted_spectra(inds, weights)

    def combine_weighted_spectra(self, inds, weights):
        """Weight self.spectra using broadcasting, then sum the
        weighted spectra.  Should add a switch to sum the weights
        first so that the output is a single spectrum.

        :param inds: shape (ntarg,nint)
            Indices of the models to sum

        :param weights: shape (ntarg, nint)
            Weights of the models corresponding to inds.

        :returns spec: shape (nwave,ntarg).
            The weighted sum.
        """
        tmp = weights * (self.spectra[inds].transpose(2, 0, 1))
        return (tmp.sum(axis=2)).T

    def read_model_from_fitsbinary(self, filename, parnames,
                                   wavename='WAVE', fluxname='F_LAMBDA'):
        """Read spectra from a fits binary table in a certain format.
        """
        # if os.ispath(filename) is False:
        #      raise IOError('read_model_from_fitsbinary:
        #                    ',filename,' does not exist')
        fits = pyfits.open(filename)
        # Parse the FITS recarray and assign ModelGrid parameter,
        # spectra, and wavelength attributes
        wavelength = fits[1].data[0][wavename]
        spectra = fits[1].data[fluxname]  # (nmod,nwave)
        # Need to loop over pars and restruct to get the desired
        # order.  Really?  Not really.  order is unimportant
        pars, partype = [], []
        for pname in parnames:
            pars.append(np.squeeze(fits[1].data[pname]))
            partype.append(fits[1].data[pname].dtype)  # fix
        # list ->(nmod, npar) -> Structured array
        pars = self.structure_array(np.array(pars).transpose(),
                                    parnames, types=partype)
        fits.close()
        return wavelength, spectra, pars



# -*- coding: utf-8 -*-

import numpy as np
from numpy.testing import assert_allclose

from sedpy import observate


def _get_all_filters():
    fnames = observate.list_available_filters()
    allfilters = observate.load_filters(fnames)
    return allfilters


def test_filter_load():
    fnames = observate.list_available_filters()
    for fn in fnames:
        try:
            f = observate.Filter(fn)
        except(IOError, ValueError, IndexError, AssertionError) as e:
            print(f"Could not build `{fn}`")
            raise(e)
        w = f.wave_effective
        assert np.isfinite(w), f"effective wavelength undefined for {fn}"


def test_gridded_filters():
    allfilters = np.array(_get_all_filters())
    w = np.array([f.wave_effective for f in allfilters])
    fnames = np.array([f.name for f in allfilters])

    good = w < 1e5
    obs = {}
    obs['filters'] = allfilters[good][0:40]

    spec = np.random.uniform(0, 1.0, 5996)
    wave = np.exp(np.linspace(np.log(90), np.log(1e6), len(spec)))
    #%timeit m = observate.getSED(wave, spec, obs['filters'])
    #3.22 ms
    m_default = observate.getSED(wave, spec, obs['filters'])

    wlo, whi, dlo = [], [], []
    for f in obs['filters']:
        dlnlam = np.gradient(f.wavelength)/f.wavelength
        wlo.append(f.wavelength.min())
        dlo.append(dlnlam.min())
        whi.append(f.wavelength.max())
    wmin = np.min(wlo)
    wmax = np.max(whi)
    dlnlam = np.min(dlo)
    obs['filters'] = observate.load_filters(fnames[good][0:40],
                                            dlnlam=dlnlam, wmin=wmin)

    lnlam = np.exp(np.arange(np.log(wmin), np.log(wmax), dlnlam))
    lnspec = np.interp(lnlam, wave, spec)

    m_grid = observate.getSED(lnlam, lnspec, obs['filters'],
                              gridded=True)
    # make sure good to 5% (most far better, hipparcos_B is a problem)
    assert np.allclose(m_grid, m_default, atol=5e-2)


def test_filterset():
    from sedpy.observate import FilterSet
    flist = [f for f in observate.list_available_filters()
             if ("jw" in f) or ("wise" in f) or ("galex" in f)]
    filterset = FilterSet(flist, dlnlam=1e-4)
    source = observate.vega.T

    mags = observate.getSED(source[0], source[1], filterset)
    omags = observate.getSED(source[0], source[1], observate.load_filters(flist))

    good = np.isclose(mags, omags, atol=1.0e-3)
    assert np.all(good), f"filters {np.array(flist)[~good]} failed with {(mags-omags)[~good]}"


def test_shapes():

    from sedpy.observate import FilterSet
    flist = [f for f in observate.list_available_filters()
             if ("jw" in f) or ("wise" in f) or ("galex" in f)]
    sw, sf0 = observate.vega.T
    assert sf0.ndim == 1

    for fl in [flist[:1], flist]:
        Nf = len(fl)
        filterset = FilterSet(fl)

        # test input flux of shape (nwave,)
        mags = observate.getSED(sw, sf0, filterset)
        assert mags.shape == (Nf,), "getSED with FilterSet gives {}, expected {}".format(mags.shape, (Nf,))
        omags = observate.getSED(sw, sf0, observate.load_filters(fl))
        assert omags.shape == (Nf,), "getSED with Filter list gives {}, expected {}".format(omags.shape, (Nf,))

        # test vector inputs, though inputs of shape (1, nwave) get turned into (nfilt,)
        for N in [1, 10]:
            if N == 1:
                truth = (Nf,)
            else:
                truth = (N, Nf)

            sf = np.tile(sf0, (N, 1))
            mags = observate.getSED(sw, sf, filterset)
            assert mags.shape == truth, "getSED with FilterSet gives {}, expected {}".format(mags.shape, truth)
            omags = observate.getSED(sw, sf, observate.load_filters(fl))
            assert omags.shape == truth, "getSED with Filter list gives {}, expected {}".format(omags.shape, truth)

            assert np.allclose(mags, omags, 5e-3), mags - omags


def test_gridded_shapes():

    from sedpy.observate import FilterSet

    # array of object spectra
    Nobj, Nwave = 100, 3000
    spec = np.ones([Nobj, Nwave], dtype=float)
    wave = np.exp(np.linspace(np.log(0.5e4), np.log(5.1e4), Nwave))

    fnames = [f"jwst_f{b}" for b in ["070w", "090w", "115w", "150w", "200w", "335m"]]
    fnames.sort()

    wmin = wave.min()
    dlnlam = np.gradient(np.log(wave))
    dlnlam_filters = dlnlam.min()
    assert np.allclose(dlnlam, dlnlam[0])

    filterset = FilterSet(fnames, dlnlam=dlnlam_filters, wmin=wmin)
    inds = np.log(wave) <= (np.log(filterset.lam.max()) + dlnlam_filters*0.1)

    maggies = filterset.get_sed_maggies(spec[:, inds])
    assert maggies.shape == (100, len(fnames))


def test_filter_properties():
    """Compare to the values obtained from the K-correct code
    (which uses a slightly different Vega spectrum)
    """
    filternames = ['galex_FUV',
                   'sdss_u0', 'sdss_g0', 'sdss_r0', 'sdss_i0',
                   'spitzer_irac_ch2']
    weff_kcorr = [1528.0,
                  3546.0, 4669.6, 6156.25, 7471.57,
                  44826.]
    msun_kcorr = [18.8462,
                  6.38989, 5.12388, 4.64505, 4.53257,
                  6.56205]
    ab2vega_kcorr = [2.3457,
                     0.932765, -0.0857, 0.155485, 0.369598,
                     3.2687]

    filterlist = observate.load_filters(filternames)
    for i, f in enumerate(filterlist):
        #print(filterlist[i].wave_effective, filterlist[i].solar_ab_mag,
        #      filterlist[i].ab_to_vega)
        assert (abs(f.wave_effective - weff_kcorr[i]) <
                (weff_kcorr[i] * 0.01))
        assert abs(f.solar_ab_mag - msun_kcorr[i]) < 0.05
        # the below fails because of the vega spectrum used by k_correct
        # assert abs(f.ab_to_vega+ab2vega_kcorr[i]) < 0.05



# -- VERY PRELIMINARY!!!! -----

import numpy as np
try:
    from matplotlib import path
except(ImportError):
    pass


region_types = ['circle', 'ellipse', 'polygon', 'box']


def load_regfile(regfile):
    reglist = []
    f = open(regfile, 'r')
    for line in f:
        istype = [r in line for r in region_types]
        if True in istype:
            rtype = region_types[np.where(istype)[0][0]]
            numstring = line[line.find("(")+1:line.find(")")].replace('"', '')
            if rtype == 'ellipse':
                reglist.append(Ellipse(numstring))
            elif rtype == 'circle':
                reglist.append(Circle(numstring))
            elif rtype == 'polygon':
                reglist.append(Polygon(numstring))
    f.close()
    return reglist


def get_image_angle(wcs, at_coord=None):
    """
    :returns angle:
       In radians, degrees East of North
    """
    if at_coord is None:
        lon, lat = wcs.wcs.crval[:2]
    else:
        lon, lat = at_coord
    x, y = world2pix(wcs, lon, lat)
    lat_offset = lat + 1./3600.
    xoff, yoff = world2pix(wcs, lon, lat_offset)
    angle = np.arctan2(xoff-x, yoff-y)
    try:
        return angle[0]
    except(TypeError, IndexError):
        return angle


def world2pix(wcs, lon, lat, **extras):
    """Wrap the wcs world2pix method to be less finicky and take our defaults.
    """
    try:
        x, y = wcs.wcs_world2pix(lon, lat, 0)
    except(TypeError):
        x, y, _ = wcs.wcs_world2pix(lon, lat, np.array([0]), 0)
    return x, y


class Region(object):

    def __init__(self, defstring):
        self.parse(defstring)

    def parse(self, defstring):
        raise(NotImplementedError)

    @property
    def unparse(self):
        raise(NotImplementedError)

    def plate_scale(self, wcs):
        try:
            pscale = 3600. * np.sqrt((wcs.wcs.cd[:2, :2]**2).sum(axis=1))
        except (AttributeError):
            pscale = 3600. * np.abs(wcs.wcs.cdelt[:2])
        return pscale.mean()

    def print_to(self, fileobj=None, color='green', label=''):
        fstring = 'fk5;{0}({1}) # color={2} text={{{3}}}\n'
        line = fstring.format(self.shape, self.unparse, color, label)
        if fileobj is None:
            print(line)
        else:
            fileobj.write(line)


class Circle(Region):

    shape = 'circle'

    def parse(self, defstring):
        bits = defstring.split(',')
        self.ra = float(bits[0])
        self.dec = float(bits[1])
        self.radius = float(bits[2])

    @property
    def unparse(self):
        return ','.join([str(a) for a in [self.ra, self.dec, self.radius]])

    def contains(self, points=None, x=None, y=None, wcs=None, **extras):
        if wcs is not None:
            # Should actually do the plate scale separately for x and y
            # and use great circle distances?
            plate_scale = self.plate_scale(wcs)
            r = self.radius / plate_scale
            cx, cy = world2pix(wcs, self.ra, self.dec)
        else:
            raise ValueError('No WCS given!!!')
        if points is not None:
            x, y = points.T

        dx, dy = x-cx, y-cy
        return (dx**2 + dy**2) <= r**2


class Ellipse(Region):

    shape = 'ellipse'

    def parse(self, defstring):
        bits = defstring.split(',')
        self.ra = float(bits[0])   #degrees
        self.dec = float(bits[1])  #degrees
        self.a = float(bits[2])    #arcsec
        self.b = float(bits[3])    #arcsec
        self.pa = float(bits[4])   #degrees East of North for a

        #swap a and b if poorly defined
        if self.b > self.a:
            self.a, self.b = self.b, self.a
            self.pa = self.pa + 90.

    @property
    def unparse(self):
        return ','.join([str(a) for a in [self.ra, self.dec, self.a, self.b, self.pa]])

    def contains(self, points=None, x=None, y=None, wcs=None, **extras):
        """
        Determine whether pixel locations x,y are within the ellipse.
        Requires that a WCS be given.  Assumes tangent projection, and
        will fail for large ellipses/images. Assumes N is up and E to
        the left.

        x, and y are zero indexed, and can be produced by flattened versions of
        `y, x = np.indices(image)`
        """
        if wcs is not None:
            cx, cy = world2pix(wcs, self.ra, self.dec)
            # should actually do the plate scale separately for x and y
            # and use great circle distances?
            plate_scale = self.plate_scale(wcs)
            a, b = self.a / plate_scale, self.b / plate_scale
            theta_image = get_image_angle(wcs)
        else:
            raise ValueError('No WCS given!!!')
        if points is not None:
            x, y = points.T

        # --- Translate ---
        dx, dy = x - cx, y - cy
        # --- Rotate ---
        #  convert to radians from positive x
        theta = np.deg2rad(self.pa) - theta_image + np.pi
        # construct rotation matrix
        R = np.array([[np.cos(theta), -np.sin(theta)],
                      [np.sin(theta), np.cos(theta)]])
        # We flip dx and dy here because why?
        # it seems to work.... something to do with column-major/row-major
        prime = np.dot(R, np.vstack([dy, dx]))

        r = np.hypot(prime[0, :] / a, prime[1, :] / b)
        #side1 = ((dx * np.cos(np.deg2rad(self.pa-90)) - dy * np.sin(np.deg2rad(self.pa-90))) / a)**2
        #side2 = ((dx * np.sin(np.deg2rad(self.pa-90)) + dy * np.cos(np.deg2rad(self.pa-90))) / b)**2
        return r <= 1


class Polygon(Region):

    shape = 'polygon'

    def parse(self, defstring):
        bits = defstring.split(',')
        bits = np.array([float(b) for b in bits])
        #print( len(bits), bits)
        assert (len(bits) % 2) == 0
        self.n_vertices = len(bits)/2
        self.ra = bits[np.arange(self.n_vertices) * 2]       #degrees
        self.dec = bits[np.arange(self.n_vertices) * 2 + 1]  #degrees

    @property
    def unparse(self):
        return ','.join([str(val) for pair in zip(self.ra, self.dec) for val in pair])

    def contains(self, points=None, x=None, y=None, wcs=None,
                 fast=False, pad=[0, 0], **extras):
        """
        Determine whether pixel locations x,y are within the polygon.
        Requires that a WCS be given.  Uses Path objects from matlib,
        should probably use Shapely, or better yet a real spherical
        coordinates implementation
        """
        if wcs is not None:
            vv = world2pix(wcs, self.ra, self.dec)
            #vv[0] -= 1 #convert to numpy indexing
        else:
            vv = (self.ra, self.dec)

        vertices = [(r, d) for r, d in zip(vv[0], vv[1])]
        vertices += [vertices[0]]
        poly = path.Path(vertices, closed=True)
        if points is None:
            points = np.vstack((x, y)).T

        if fast:
            sub = ((points[:, 0] > vv[0].min()-pad[0]) &
                   (points[:, 0] < vv[0].max()+pad[0]) &
                   (points[:, 1] > vv[1].min()-pad[1]) &
                   (points[:, 1] < vv[1].max()+pad[1]))
            sub_is_in = poly.contains_points(points[sub, :])
            is_in = np.zeros(points.shape[0], dtype=bool)
            is_in[sub] = sub_is_in
        else:
            is_in = poly.contains_points(points)
            #flux = (self.image - background)[is_in].sum()
        return is_in



# -*- coding: utf-8 -*-

"""observate.py - Python module for storing filter information and tools for
projecting spectra onto filters.  Also includes tools for convolving spectra.

Assumed input units are erg/s/cm^2/AA and AA
"""

import numpy as np
import os
try:
    from pkg_resources import resource_filename, resource_listdir
except(ImportError):
    pass
from .reference_spectra import vega, solar, sedpydir


__all__ = ["Filter", "FilterSet",
           "load_filters", "list_available_filters", "getSED",
           "air2vac", "vac2air", "Lbol"]


lightspeed = 2.998e18  # AA/s


class Filter(object):
    """This class operates on filter transmission files.  It reads simple
    2-column ascii files containing filter transmission curves. It caches a
    number of useful filter quantities.  Methods are provided to project a
    source spectrum onto the filter and return the magnitude, AB or Vega.

    Parameters
    ----------
    kname : string (default: 'sdss_r0')
        The kcorrect style name of the filter, excluing '.par',
        e.g. sdss_r0.

    nick : string (optional)
        A nickname to associate with the filter.

    directory : string (optional)
        The path to the directory containing the filter file.  If not given
        then the sedpy/data/filters/ directory will be searched.

    dlnlam : float (optional)
        If given, interpolate the transmission curve onto a grid in
        :math:`\ln(\lambda)` with this spacing (and begining at ``wmin``)

    wmin : float (optional, default 100.)
        If ``dlnlam`` is supplied then ``wmin`` gives the minimum wavelength
        (in Angstroms) for the grid.

    min_trans : float (optional, default 1e-5)
        The minimum transmission value (as fraction of the maximum transmission
        value) to consider as a useful positive value.  Useful for curves that
        have extremely low amplitude components far from the main part of the
        filter when speed is more important than accuracy.

    data: None or tuple of ndarray (optional, default None)
        If provided, a 2-tuple of ndarrays giving wavelength (AA) and
        transmission (detector signal per photon).
    """
    ab_gnu = 3.631e-20  # AB reference spctrum in erg/s/cm^2/Hz
    npts = 0

    def __init__(self, kname='sdss_r0', nick=None, directory=None,
                 dlnlam=None, wmin=1e2, min_trans=1e-5, data=None, **extras):
        """Read in the filter data, calculate and cache some properties of the
        filters.
        """
        self.name = kname
        if nick is None:
            self.nick = kname
        else:
            self.nick = nick

        self.min_trans = min_trans

        if directory is None:
            loc = os.path.join('data', 'filters', kname + '.par')
            try:
                self.filename = resource_filename("sedpy", loc)
            except:
                self.filename = os.path.join(sedpydir, loc)
        else:
            self.filename = os.path.join(directory, kname+'.par')

        # Load wave, trans directly from data arrays
        if data is not None:
            wave, trans = data
            self._process_filter_data(wave, trans)
            self.filename = None

        if isinstance(self.filename, str):
            if not os.path.isfile(self.filename):
                raise ValueError('Filter transmission file {0} does '
                                 'not exist!'.format(self.filename))
            self.load_filter(self.filename)
        elif isinstance(self.filename, type(None)):
            pass
        else:
            msg = "{} of type {} is not a valid transmission file name: check your string types?"
            raise TypeError(msg.format(self.filename, type(self.filename)))

        if dlnlam is not None:
            self.gridify_transmission(dlnlam, wmin=wmin)

        self.get_properties()

    def __repr__(self):
        return '{}({})'.format(self.__class__, self.name)

    def load_filter(self, filename):
        """Read a filter in simple two column ascii format and populate the
        wavelength and transmission arrays.  The first column is wavelength in
        AA and the second column is transmission (detector signal per photon)

        Parameters
        ----------
        filename : string
            The fully qualified path and filename of the file that contains the
            filter transmission.
        """
        wave, trans = np.genfromtxt(filename, usecols=(0, 1), unpack=True)
        # Clean negatives, NaNs, and Infs, then sort, then store
        self._process_filter_data(wave, trans)

    @property
    def transmission(self):
        return self._transmission

    @property
    def wavelength(self):
        return self._wavelength

    def _process_filter_data(self, wave, trans):
        """Clean up transmission data and assign to attributes.

        Parameters
        ----------
        wave : ndarray of shape ``(N_wave)``
            Wavelength, in Angstroms, of the transmission curve

        trans : ndarray of shape ``(N_wave)``
            Filter transmission at the wavelengths corresponding to ``wave``
        """
        ind = np.isfinite(trans) & (trans >= 0.0)
        order = wave[ind].argsort()
        self.npts = ind.sum()
        self._wavelength = wave[ind][order]
        self._transmission = trans[ind][order]
        self._remove_extra_zeros(self.min_trans)

    def _remove_extra_zeros(self, min_trans=0):
        """Remove extra leading or trailing zero transmission points.  This
        leaves one zero before (after) the first (last) non-zero transmission
        point, if they were present in the original transmission function.

        Parameters
        ----------
        min_trans : float (optional, default: 1e-5)
            Defines zero, in terms of fraction of the maximum transmission.
        """
        v = np.argwhere(self.transmission > (self.transmission.max() * min_trans))
        inds = slice(max(v.min() - 1, 0), min(v.max() + 2, len(self.transmission)))
        self._wavelength = self._wavelength[inds]
        self._transmission = self._transmission[inds]
        self.npts = len(self.wavelength)

    def gridify_transmission(self, dlnlam, wmin=1e2):
        """Place the transmission function on a regular grid in lnlam
        (angstroms) defined by a lam_min and dlnlam.  Note that only the
        non-zero values of the transmission on this grid stored.  The indices
        corresponding to these values are stored as the `inds` attribute (a
        slice object). (with possibly a zero at either end.)

        Parameters
        ----------
        dlnlam : float
            The spacing in ln-lambda of the regular wavelength grid onto which
            the filter is to be placed.

        wmin : float (optional, default: 100)
            The starting wavelength (Angstroms) for the regular grid.
        """
        # find min and max of the filter in a global wavelength grid given by
        # wmin, wmax, and dlnlam
        ind_min = int(np.floor((np.log(self.wavelength.min()) - np.log(wmin)) / dlnlam))
        ind_max = int(np.ceil((np.log(self.wavelength.max()) - np.log(wmin)) / dlnlam))
        lnlam = np.linspace(ind_min * dlnlam + np.log(wmin),
                            ind_max * dlnlam + np.log(wmin), ind_max - ind_min)
        lam = np.exp(lnlam)
        # TODO: replace with a rebinning
        trans = np.interp(lam, self.wavelength, self.transmission,
                          left=0., right=0.)

        self.wmin = wmin
        self.dlnlam = dlnlam
        self.inds = slice(ind_min, ind_max)
        self._wavelength = lam
        self._transmission = trans
        self.dwave = np.gradient(lam)

    def get_properties(self):
        """Determine and store a number of properties of the filter and store
        them in the object.  These properties include several 'effective'
        wavelength definitions and several width definitions, as well as the
        in-band absolute AB solar magnitude, the Vega and AB reference
        zero-point detector signal, and the conversion between AB and Vega
        magnitudes.

        See Fukugita et al. (1996) AJ 111, 1748 for discussion and definition
        of many of these quantities.
        """
        # Calculate some useful integrals
        i0 = np.trapz(self.transmission * np.log(self.wavelength),
                      np.log(self.wavelength))
        i1 = np.trapz(self.transmission,
                      np.log(self.wavelength))
        i2 = np.trapz(self.transmission * self.wavelength,
                      self.wavelength)
        i3 = np.trapz(self.transmission,
                      self.wavelength)

        tmax = self.transmission.max()
        self.wave_effective = np.exp(i0 / i1)
        self.wave_pivot = np.sqrt(i2 / i1)
        self.wave_mean = self.wave_effective
        self.wave_average = i2 / i3
        self.rectangular_width = i3 / tmax

        wpeak = self.wavelength[np.argmax(self.transmission)]
        sel = (self.wavelength < wpeak) & (self.transmission < 0.5 * tmax)
        bind = np.argmax(self.wavelength[sel])
        self.blue_edge = self.wavelength[sel][bind]
        sel = (self.wavelength > wpeak) & (self.transmission < 0.5 * tmax)
        rind = np.argmin(self.wavelength[sel])
        self.red_edge = self.wavelength[sel][rind]

        i4 = np.trapz(self.transmission *
                      (np.log(self.wavelength / self.wave_effective))**2.0,
                      np.log(self.wavelength))
        self.gauss_width = (i4 / i1)**(0.5)
        self.effective_width = (2.0 * np.sqrt(2. * np.log(2.)) *
                                self.gauss_width *
                                self.wave_effective)
        # self.norm  = np.trapz(transmission,wavelength)

        # Get zero points and AB to Vega conversion
        self.ab_zero_counts = self.obj_counts(self.wavelength,
                                              self.ab_gnu * lightspeed /
                                              self.wavelength**2)
        # If blue enough get AB mag of vega
        if self.wave_mean < 1e6:
            self.vega_zero_counts = self.obj_counts(vega[:, 0], vega[:, 1])
            self._ab_to_vega = -2.5 * np.log10(self.ab_zero_counts /
                                               self.vega_zero_counts)
        else:
            self.vega_zero_counts = float('NaN')
            self._ab_to_vega = float('NaN')
        # If blue enough get absolute solar magnitude
        if self.wave_mean < 1e5:
            self.solar_ab_mag = self.ab_mag(solar[:,0], solar[:,1])
        else:
            self.solar_ab_mag = float('NaN')

    @property
    def ab_to_vega(self):
        """The conversion from AB to Vega systems for this filter.  It has the
        sense

        :math:`m_{Vega} = m_{AB} + Filter().ab_to_vega`
        """
        return self._ab_to_vega

    def display(self, normalize=False, ax=None):
        """Plot the filter transmission curve
        """
        if self.npts > 0:
            if ax is None:
                import matplotlib.pyplot as pl
                _, ax = pl.subplots()
                ax.set_title(self.nick)
            if normalize:
                ax.plot(self.wavelength, self.transmission / self.transmission.max())
            else:
                ax.plot(self.wavelength, self.transmission)
            return ax

    def obj_counts_hires(self, sourcewave, sourceflux, sourceflux_unc=0):
        """Project source spectrum onto filter and return the detector signal.

        Parameters
        ----------
        sourcewave : ndarray of shape ``(N_pix,)``
            Spectrum wavelength (in Angstroms). Must be monotonic increasing.

        sourceflux : ndarray of shape ``(N_source, N_pix)``
            Associated flux (assumed to be in erg/s/cm^2/AA)

        Returns
        -------
        counts : ndarray of shape ``(N_source,)``
            Detector signal(s).
        """
        assert sourcewave[1] > sourcewave[0], "``sourcewave`` not in ascending order."
        # Interpolate filter transmission to source spectrum
        newtrans = np.interp(sourcewave, self.wavelength, self.transmission,
                             left=0., right=0.)

        # Integrate lambda*f_lambda*R
        if True in (newtrans > 0.):
            positive = np.where(newtrans > 0.)[0]
            edge = (positive.min() < 1) | (positive.max() >= (len(newtrans)-1))
            # assert ~edge, "Source spectrum does not span filter."
            ind = slice(max(positive.min() - 1, 0),
                        min(positive.max() + 2, len(sourcewave)))
            counts = np.trapz(sourcewave[ind] * newtrans[ind] *
                              sourceflux[..., ind],
                              sourcewave[ind], axis=-1)
            return np.squeeze(counts)
        else:
            return float('NaN')

    def obj_counts_lores(self, sourcewave, sourceflux, sourceflux_unc=0):
        """Project source spectrum onto filter and return the detector
        signal. This method differs from ``obj_counts_hires`` in that the
        source spectrum is interpolated onto the transmission spectrum instead
        of vice-versa, which is necessary with the source spectrum does not
        adequately sample features in the transmission spectrum.

        Parameters
        ----------
        sourcewave : ndarray of shape ``(N_pix,)``
            Spectrum wavelength (in Angstroms). Must be monotonic increasing.

        sourceflux : ndarray of shape ``(N_source, N_pix)``
            Associated flux (assumed to be in erg/s/cm^2/AA)

        Returns
        -------
        counts : ndarray of shape ``(N_source,)``
            Detector signal(s).
        """
        sourceflux = np.squeeze(sourceflux)
        assert sourceflux.ndim == 1, "Only a single source allowed."
        assert sourcewave[1] > sourcewave[0], "``sourcewave`` not in ascending order."
        # Interpolate source spectrum to filter transmission
        newflux = np.interp(self.wavelength, sourcewave, sourceflux,
                            left=0., right=0.)

        # Integrate lambda*f_lambda*R
        if True in (newflux > 0.):
            counts = np.trapz(self.wavelength * self.transmission * newflux,
                              self.wavelength)
            return np.squeeze(counts)
        else:
            return float('NaN')

    def obj_counts_grid(self, sourceflux, source_offset=None):
        """Project source spectrum onto filter and return the detector
        signal. This method differs from ``obj_counts_*res`` in that the source
        spectrum is assumed to be interpolated onto a logarithmic grid in
        lambda with spacing and minimum wavelength given by `dlnlam` and `wmin`
        filter attributes.

        Parameters
        ----------
        sourceflux : ndarray of shape ``(N_source, N_pix)``
            Source flux (assumed to be in erg/s/cm^2/AA) assumed to be on the
            logarithmic grid defined by ``wmin`` and ``dlnlam``, with wavelength
            ascending.

        source_offset : int (optional)
            The (hypothetical) element of the sourceflux array corresponding to
            observed frame wavelength of ``wmin``.  Should be negative if the first
            element of sourcewave corresponds to a wavelength > wmin.

        Returns
        -------
        counts : ndarray of shape ``(N_source,)``
            Detector signal(s).
        """
        # Note that redshifting can also be incorporated using negative `source_offset`,
        # modulo factors of (1+z).  I think.

        if source_offset is None:
            valid = self.inds
        else:
            valid = slice(self.inds.start + source_offset,
                          self.inds.stop + source_offset)
        counts = np.sum(sourceflux[..., valid] * self.transmission *
                        self.wavelength * self.dwave, axis=-1)
        return np.squeeze(counts)

    def obj_counts(self, sourcewave, sourceflux, lores=False, gridded=False, **extras):
        """Project a spectrum onto a filter and return the detector signal.
        This method uses the keywords `lores` and `gridded` to choose between
        the various projection algorithms.

        Parameters
        ----------
        sourcewave : ndarray of shape ``(N_pix,)``
            Spectrum wavelength (in Angstroms). Must be monotonic increasing.

        sourceflux : ndarray of shape ``(N_source, N_pix)``
            Associated flux (assumed to be in erg/s/cm^2/AA)

        lores : bool (optional, default: False)
            Switch to interpolate the source spectrum onto the wavelength grid
            of the filter transmission curve, instead of the vice-versa (which
            is the default).  Useful for a spectrum with sparse sampling in
            wavelength compared to the filter transmission curve (e.g. narrow
            bands and BaSeL lores spectra)

        gridded : bool (optional, default: False)
            Switch to accomplish the filter projection via simple sums.  This
            can be faster, but assumes the sourcewave and sourceflux are on a
            logarithmic wavelength grid given by the dlnlam and wmin attributes
            of the Filter object.

        Returns
        -------
        counts : ndarray of shape ``(N_source,)``
            Detector signal(s).

        """
        if gridded:
            counts = self.obj_counts_grid(sourceflux, **extras)
        elif lores:
            counts = self.obj_counts_lores(sourcewave, sourceflux, **extras)
        else:
            counts = self.obj_counts_hires(sourcewave, sourceflux, **extras)

        return counts

    def ab_mag(self, sourcewave, sourceflux, **extras):
        """Project source spectrum onto filter and return the AB magnitude.

        Parameters
        ----------
        sourcewave : ndarray of shape ``(N_pix,)``
            Spectrum wavelength (in Angstroms). Must be monotonic increasing.

        sourceflux : ndarray of shape ``(N_source, N_pix)``
            Associated flux (assumed to be in erg/s/cm^2/AA)

        Returns
        -------
        mag : float or ndarray of shape ``(N_source,)``
            AB magnitude of the source(s).
        """
        counts = self.obj_counts(sourcewave, sourceflux, **extras)
        return -2.5 * np.log10(counts / self.ab_zero_counts)

    def vega_mag(self, sourcewave, sourceflux, **extras):
        """Project source spectrum onto filter and return the Vega magnitude.

        Parameters
        ----------
        sourcewave : ndarray of shape ``(N_pix,)``
            Spectrum wavelength (in Angstroms). Must be monotonic increasing.

        sourceflux : ndarray of shape ``(N_source, N_pix)``
            Associated flux (assumed to be in erg/s/cm^2/AA)

        Returns
        -------
        mag : float or ndarray of shape ``(N_source,)``
            Vega magnitude of the source(s).
        """
        counts = self.obj_counts(sourcewave, sourceflux, **extras)
        return -2.5 * np.log10(counts / self.vega_zero_counts)


class FilterSet(object):
    """
    A class representing a set of filters, and enabling fast projections of
    sources onto these filters via pre-computation, caching, and a global
    wavelength array.
    """

    def __init__(self, filterlist, wmin=None, dlnlam=None,
                 **loading_kwargs):
        """
        Parameters
        ----------
        filterlist : list of strings
            The filters to include in this set, in order

        wmin : float
            Minimum wavelength (AA) for the wavelength grid for the FilterSet

        dlnlam : float
            logarithmic spacing for the wavelength grid of the FilterSet
        """
        self.filternames = filterlist
        native = load_filters(self.filternames, **loading_kwargs)
        self._set_filters(native, wmin=wmin, dlnlam=dlnlam, **loading_kwargs)
        self._build_super_trans()

    def __len__(self):
        return len(self.filternames)

    def _set_filters(self, native, wmin=None, wmax=None, dlnlam=None,
                     **loading_kwargs):
        """Set the filters and the wavelength grid

        native : list of Filter() instances
            The Filter objects that will be part of this FilterSet, including
            their native transmission
        """
        # get the wavelength grid parameters
        self.dlnlam_native = np.array([np.diff(np.log(f.wavelength)).min()
                                       for f in native])
        if dlnlam is None:
            dlnlam = min(self.dlnlam_native.min(), 1e-3)
        if wmin is None:
            wmin = np.min([f.wavelength.min() for f in native])
        if wmax is None:
            wmax = np.max([f.wavelength.max() for f in native])
        self.wmin = wmin
        self.wmax = np.exp(np.log(wmax) + dlnlam)
        self.dlnlam = dlnlam

        # Build the wavelength grid
        self.lnlam = np.arange(np.log(self.wmin), np.log(self.wmax), self.dlnlam)
        self.lam = np.exp(self.lnlam)

        # build the filters on that grid
        self.filters = load_filters(self.filternames, wmin=self.wmin,
                                    dlnlam=self.dlnlam, **loading_kwargs)

    def _build_super_trans(self):
        """Build the global gridded transmission array. Populates FilterSet.trans, which
        is an array of shape (nfilters, nlam) giving :math:`T\times \lambda \times
        d\lambda / Z` where `T` is the detector signal per photon and `Z` is the
        detector signal for a 1 maggie source
        """
        # TODO: investigate using arrays that cover different wavelength
        # intervals (but same number of elements) for each filter. Then when
        # integrating (with dot or vectorized trapz) the source spectrum must be
        # offset for each filter. This would remove the vast majority of zeros

        ab_counts = np.array([f.ab_zero_counts for f in self.filters])

        self.trans = np.zeros([len(self.filters), len(self.lam)])
        for j, f in enumerate(self.filters):
            dwave = np.gradient(self.lam[f.inds])
            self.trans[j, f.inds] = f.transmission[:] * f.wavelength[:] * dwave / ab_counts[j]

        # this is for the numba unequal inmplementation
        self.trans_list = [self.trans[j, f.inds] for j, f in enumerate(self.filters)]
        self.frange = np.array([(f.inds.start, f.inds.stop) for f in self.filters])

    def interp_source(self, inwave, sourceflux):
        """Interpolate spectra onto the global wavelength grid.

        Parameters
        ----------
        inwave : ndarray of shape ``(N_pix,)``
            Input wavelength (Angstroms)

        sourceflux : ndarray of shape ``(N_source, N_pix)``
            Input source fluxes at the wavelengths given by `inwave`

        Returns
        -------
        flux : ndarray of shape ``(N_pix,)`` or ``(N_pix, N_source)``
           Interpolated flux. Note the transpose relative to input if ``N_source > 1``
        """
        sourceflux = np.atleast_2d(sourceflux)
        # TODO: replace with scipy interpolate for vectorization?
        # TODO replace with rebinning!!!!
        flux = np.array([np.interp(self.lam, inwave, s, left=0, right=0)
                         for s in sourceflux])
        return np.squeeze(flux.T)

    def get_sed_maggies(self, sourceflux, sourcewave=None):
        """Compute the flux in maggies of the source through the given filters.

        Parameters
        ----------
        sourceflux : ndarray of shape ``(N_source, N_pix)``
            Source flux (assumed to be in erg/s/cm^2/AA)

        sourcewave: ndarray (optional)
            If supplied, the wavelength in Angstroms.  Otherwise it is assumed
            to be on the same wavelength grid as FilterSet.lam, saving the cost
            of an interpolation.

        Returns
        -------
        maggies : ndarray of shape ``(N_filter,)`` or ``(N_source, N_filter)``
            The flux (in maggies) of the sources through the filters.
        """
        if sourcewave is not None:
            flux = self.interp_source(sourcewave, sourceflux)
        else:
            flux = sourceflux.T
            assert len(flux) == len(self.lam)
        maggies = np.dot(self.trans, flux)

        return maggies.T

    def _build_trans_chunks(self):
        # NOTE: For USE WITH vectorized TRAPZ
        ab_counts = np.array([f.ab_zero_counts for f in self.filters])
        self.fstart = np.array([f.inds.start for f in self.filters])
        self.fstop = np.array([f.inds.stop for f in self.filters])
        self.nf = self.fstop - self.fstart

        # build an index array in lnlam that produces (nfilter, nf.max()) shape array
        self.finds = self.fstart[:, None] + range(self.nf.max())
        finds_right = self.fstop[:, None] - range(self.nf.max())[::-1]
        right = self.finds.max(axis=-1) >= len(self.lam)
        self.finds[right, :] = finds_right[right, :]

        # fill the transmission arrays, shifting some to the right (i.e. zero padding on the left)
        # to make them fit.  Note this is T \times \lambda, but *not* times dw
        self.trans_dw = np.zeros([len(self.filters), self.nf.max()])
        for j, f in enumerate(self.filters):
            #dwave = np.gradient(self.lam[f.inds])
            t = f.transmission[:] * f.wavelength[:] / ab_counts[j]
            if right[j]:
                self.trans_dw[j, -self.nf[j]:] = t
            else:
                self.trans_dw[j, :self.nf[j]] = t

    def get_sed_maggies_trapz(self, sourceflux, sourcewave=None):
        """Use vectorized trapz to project filters.  This can be
        slower than the default get_sed_maggies
        """
        # NOTE: THIS IS SLOWER THAN GET_SED_MAGGIES
        if sourcewave is not None:
            flux = self.interp_source(sourcewave, sourceflux)
        else:
            assert len(sourceflux) == len(self.lam)
            flux = sourceflux
        y = self.trans_dw * flux[self.finds]
        x = self.lam[self.finds]
        maggies = np.trapz(y, x, axis=-1)
        return maggies

    def get_sed_maggies_numba(self, sourceflux, sourcewave=None):
        if sourcewave is not None:
            source = self.interp_source(sourcewave, sourceflux)
        else:
            source = sourceflux

        # Allow for uneven transmission curve vectors
        maggies = _project_filters_fast(len(self), self.frange, self.trans_list, source)

        return maggies

try:
    import numba
    @numba.njit
    def _project_filters_fast(nfilt, filter_inds,
                              trans_list, source_vector):
        maggies = np.empty(nfilt)
        for i in range(nfilt):
            start, stop = filter_inds[i]
            source = source_vector[start:stop]
            maggies[i] = np.sum(trans_list[i] * source)
        return maggies

    #assert _project_filters_fast.nopython_signatures

except(ImportError):
    def _project_filters_fast(nfilt, filter_inds,
                              trans_list, source_vector):
        raise ImportError("Could not import numba")

# ------------
# Useful utilities
# -------------


def rebin(outwave, wave, trans):
    """Rebin (instead of interpolate) transmission onto given wavelength array,
    preserving total transmission.  This is useful if the output wavelength
    array is more coarseley sampled than the native transmission array, to avoid
    interpolating over important sharp features in the transmission function

    Stripped down version of specutils.FluxConservingResampler

    FIXME: implement as NUFFT?  Do something about the jaggedness (high frequency noise)
    introduced when the output is more finely sampled than the input?
    """
    assert np.all(np.diff(wave) > 0)
    assert np.all(np.diff(outwave) > 0)

    # Assume input is based on centers, so we need to work out edges
    edges = (wave[:-1] + wave[1:]) / 2
    edges = np.concatenate([[2*edges[0]-edges[1]], edges, [2*edges[-1] - edges[-2]]])
    inlo, inhi = edges[:-1], edges[1:]
    edges = (outwave[:-1] + outwave[1:]) / 2
    edges = np.concatenate([[2*edges[0]-edges[1]], edges, [2*edges[-1] - edges[-2]]])
    outlo, outhi = edges[:-1], edges[1:]

    # Make a huge matrix of the contribution of each input bin to each output bin
    # Basically a brute force convolution with variable width square kernel
    l_inf = np.where(inlo > outlo[:, None], inlo, outlo[:, None])
    l_sup = np.where(inhi < outhi[:, None], inhi, outhi[:, None])
    resamp_mat = (l_sup - l_inf).clip(0) * (inhi - inlo)

    # remove any contribution to output bins that aren't totally within the interval of the original bins.
    left_clip = np.where(outlo - inlo[0] < 0, 0, 1)
    right_clip = np.where(inhi[-1] - outhi < 0, 0, 1)
    keep_overlapping_matrix = left_clip * right_clip
    resamp_mat *= keep_overlapping_matrix[:, None]

    # Do the convolution
    out = np.dot(resamp_mat, trans) / resamp_mat.sum(axis=-1)

    return out


def load_filters(filternamelist, **kwargs):
    """Given a list of filter names, this method returns a list of Filter
    objects.

    Parameters
    ----------
    filternamelist : list of strings
        The names of the filters.

    Returns
    -------
    filterlist : list of Filter() instances
        A list of filter objects.
    """
    return [Filter(f, **kwargs) for f in filternamelist]


def getSED(sourcewave, sourceflux, filterlist=None, linear_flux=False, **kwargs):
    """Takes wavelength vector, a flux array and list of Filter objects and
    returns the SED in AB magnitudes.

    Parameters
    ----------
    sourcewave : ndarray of shape ``(N_pix,)``.
        Spectrum wavelength (in Angstroms)

    sourceflux : ndarray of shape ``(N_pix,)`` or ``(N_source, N_pix)``
        Associated flux (assumed to be in erg/s/cm^2/AA)

    filterlist : list of Filter() instances or a FilterSet() instance.
        The filters onto which to project the spectra

    linear_flux : bool, optional (default: True)
        If True return maggies, else AB mags

    Returns
    -------
    sed : ndarray of shape ``(N_source, N_filter)`` or ``(N_filter,)``
        broadband magnitudes or maggies.
    """
    if hasattr(filterlist, "get_sed_maggies"):
        maggies = filterlist.get_sed_maggies(sourceflux, sourcewave=sourcewave)
        if linear_flux:
            return maggies
        else:
            return -2.5 * np.log10(maggies)

    elif filterlist is None:
        return None

    # standard loop over filters
    sflux = np.squeeze(sourceflux)
    sedshape = sflux.shape[:-1] + (len(filterlist),)
    sed = np.zeros(sedshape)
    for i, f in enumerate(filterlist):
        sed[..., i] = f.ab_mag(sourcewave, sflux, **kwargs)
    #sed = np.atleast_1d(np.squeeze(sed))
    if linear_flux:
        return 10**(-0.4 * sed)
    else:
        return sed


def list_available_filters():
    """Return a list of filter names that are available by default, i.e. which
    have been installed in the sedpy/data/filters/ directory.
    """
    try:
        names = resource_listdir('sedpy', '/data/filters/')
    except:
        names = os.listdir(os.path.join(sedpydir, '/data/filters/'))

    parfiles = [n.replace('.par', '') for n in names if n[-4:] == '.par']
    parfiles.sort()
    return parfiles


def filter_dict(filterlist):
    fdict = {}
    for i, f in enumerate(filterlist):
        fdict[f.nick] = i
    return fdict

# ------------------
# Routines for spectra
# ------------------


def Lbol(wave, spec, wave_min=90, wave_max=1e6):
    """Calculate the bolometric luminosity of a spectrum or spectra.

    Parameters
    ----------
    wave : ndarray
       The wavelength vector of length nwave.

    spec : ndarray
       The spectra, of shape (...,nsource, nwave), in F_lambda.

    wave_min : float
       Minimum wavelength for the integral.

    max_wave : float
       Maximum wavelength for the integral

    Returns
    -------
    lbol :
       The bolometric luminosity, integrated from wave_min to wave_max.  Array
       of length (...nsource)
    """
    inds = np.where(np.logical_and(wave < wave_max, wave >= wave_min))
    return np.trapz(spec[..., inds[0]], wave[inds])


def air2vac(air):
    """Convert from in-air wavelengths to vacuum wavelengths.  Based on Allen's
    Astrophysical Quantities.

    Parameters
    ----------
    air : ndarray of shape ``(N_pix,)``
        The in-air wavelengths.

    Returns
    -------
    vac: ndarray of shape ``(N_pix,)``
        The corresponding vacuum wavelengths.
    """
    ss = 1e4 / air
    vac = air * (1 + 6.4328e-5 + 2.94981e-2 / (146 - ss**2) +
                 2.5540e-4 / (41 - ss**2))
    return vac


def vac2air(vac):
    """Convert from vacuum wavelengths to in-air wavelengths.  Follows the SDSS
    statement of the IAU standard from Morton 1991 ApJS.

    vac2air(air2vac(wave)) yields wave to within 1 part in a million over the
    optical range.

    Parameters
    ----------
    vac: ndarray of shape ``(N_pix,)``
        The vacuum wavelengths.

    Returns
    -------
    air : ndarray of shape ``(N_pix,)``
        The corresponding in-air wavelengths.
    """
    conv = (1.0 + 2.735182e-4 + 131.4182 / vac**2 + 2.76249e8 / vac**4)
    return vac / conv



import sys, os, glob
import subprocess
import numpy as np
from . import photometer
from astropy import wcs as pywcs
try:
    import astropy.io.fits as pyfits
except(ImportError):
    import pyfits


def run_command(cmd):
    """Open a child process, and return its exit status and stdout"""
    child = subprocess.Popen(cmd, shell =True, stderr = subprocess.PIPE, 
                             stdin=subprocess.PIPE, stdout = subprocess.PIPE)
    out = [s for s in child.stdout]
    w = child.wait()
    return os.WEXITSTATUS(w), out


class Image(object):

    def __init__(self, imagename):
        self.image_name = imagename
        self.image = pyfits.getdata(imagename)
        self.header = pyfits.getheader(imagename)
        self.wcs = pywcs.WCS(self.header)
        # a rough plate scale in arcseconds
        try:
            self.plate_scale = 3600. *np.sqrt((self.wcs.wcs.cd**2).sum(axis = 1))
        except (AttributeError):
             self.plate_scale = 3600. * np.abs(self.wcs.wcs.cdelt)
             
        self.mask = np.ones_like(self.image, dtype = bool)
        ny, nx = self.mask.shape  
        self.y, self.x = np.meshgrid(np.arange(ny), np.arange(nx))
        
    def sextract(self, catalog_name = 'junk.dat', **kwargs):
        """
        Run SExtractor on the image.  One should already have a default.param
        and a default.sex file laying around with most of the options set.  Howewver, any of the
        options can be set by supplying them as keywords, using a similar format
        to the commandline options.  Note that errors may be raised if the number  of
        apertures does not match the number specified in default.params.

        N.B command line changing of phot_apertures NOT WORKING
        """
        f   = self.image_name
        cmd = ['sex',f,'-CATALOG_NAME ' + catalog_name]
        for k, v in kwargs.iteritems():
            cmd[2] += ' -' + k.upper() + ' ' + v
        self.sex_command = cmd
        stat, out = run_command(' '.join(cmd))
        
        dat = read_sexcat(catalog_name)
        return dat

    def point_source_phot(self, sourcelist, shape, skypars):
        """
        Do simple aperture photomtry at a list of locations.  The list
        is ideally the record array returned by read_sexcat.
        """
        ## Initialize aperture (and background annulus)
        ap = photometer.Circular()
        ap.background = photometer.Annulus(bgtype = skypars['bgtype'])
        #ap.background = photometer.ZeroSky()
        flux = np.zeros(len(sourcelist))
        for i,s in enumerate(sourcelist):
            shape['xcen'] = s['x_image']-0.5
            shape['ycen'] = s['y_image']-0.5
            skypars['xcen'] = shape['xcen']
            skypars['ycen'] = shape['ycen']
            flux[i], dummy = ap.measure_flux(shape, self.image, skypars = skypars)
        return flux
    
    def area_phot(self, regionlist, use_mask = False):
        """
        Add up all the flux in some (large) regions of the sky defined
        by region objects. Fractional pixels not considered, so make
        your polygon big (or magnify the image and WCS).
        """
        if use_mask:
            mask = self.mask
        else:
            mask = np.ones_like(self.image, dtype = bool)
        for region in regionlist:
            mask = mask & self.region_bool_mask(region)
            
        flux = np.nansum(self.image[mask])
        
        return  flux

    def region_bool_mask(self, region):
        
        ny, nx = self.mask.shape        
        x, y = self.x.flatten(), self.y.flatten()
        inds = region.contains(x, y, wcs = self.wcs)
        grid = inds.reshape((ny,nx)).T
        return grid

    
def read_sexcat(filename):
    """Read a SExtractor results catalog, and return a numpy record array.
    Because astropy Tables are annoying.
    """
    f = open(filename,'r')
    colname = [] #parameter header
    skiprows = 0
    oldpos = 0
    for line in f:
        if line[0] !='#':
            ncol = len(line.split())
            break
        else:
            #parse the column descriptor line
            d = line.split()
            pos, name, units = int(d[1]), d[2].lower(),  d[-1][1:-1]
            #deal with vectors
            off = 1
            while pos > oldpos + off:
                if off == 1:
                    colname[-1] += '_1'
                colname.append(oldname + '_' + str(off+1))
                off += 1
            colname.append(name)
            oldpos, oldname = pos, name
            skiprows +=1
        
    f.close()
    dt = len(colname) * ['<f8']
    dtype = np.dtype(zip(colname, dt))
    dat = np.loadtxt(filename, dtype = np.dtype(zip(colname, dt)), skiprows = skiprows)
    return dat


    
def xylist(sourcelist, outname, physical = False, wcs = None):
    cols = ['alpha_j2000', 'delta_j2000']
    xx, yy = sourcelist[cols[0]], sourcelist[cols[1]]
    out = open(outname, 'w')
    out.write('global color=green font="helvetica 10 normal" select=1 edit=1 move=1 delete=1 include=1 fixed=0 source\n')
    for x,y in zip(xx, yy):
        out.write('fk5;circle({0},{1},5.0")\n'.format(x,y))
    out.close()
    



# -*- coding: utf-8 -*-

"""reference_spectra.py  - load and store Vega and Solar reference spectra.

Units of the reference spectra are AA and erg/s/cm^2/AA
"""

import numpy as np
import os
try:
    from pkg_resources import resource_filename
except ImportError:
    pass
try:
    import astropy.io.fits as pyfits
except ImportError:
    import pyfits

__all__ = ["vega", "solar", "sedpydir"]

# ----------
# Load useful reference spectra
# ----------

sedpydir, f = os.path.split(__file__)
sedpydir = sedpydir
try:
    vega_file = resource_filename('sedpy', 'data/alpha_lyr_stis_005.fits')
except:
    vega_file = os.path.join(sedpydir, 'data', 'alpha_lyr_stis_005.fits')

# This file should be in AA and erg/s/cm^2/AA
if os.path.isfile(vega_file):
    fits = pyfits.open(vega_file)
    vega = np.column_stack((fits[1].data.field('WAVELENGTH'),
                            fits[1].data.field('FLUX')))
    fits.close()
else:
    raise ValueError('Could not find Vega '
                     'spectrum at {0}'.format(vega_file))

try:
    solar_file = resource_filename('sedpy', 'data/sun_kurucz93.fits')
except:
    solar_file = os.path.join(sedpydir, 'data', 'sun_kurucz93.fits')

# conversion to d=10 pc from 1 AU
rat = (1.0 / (3600 * 180 / np.pi * 10))**2.0

# This file should be in AA and erg/s/cm^2/AA at 1AU
if os.path.isfile(solar_file):
    fits = pyfits.open(solar_file)
    solar = np.column_stack((fits[1].data.field('WAVELENGTH'),
                             fits[1].data.field('FLUX') * rat))
    fits.close()
else:
    raise ValueError('Could not find Solar '
                     'spectrum at {0}'.format(solar_file))



# -*- coding: utf-8 -*-

"""attenuation.py  - A collection of simple functions implementing various
proposed attenuation curves.
"""

import numpy as np
import warnings, sys

# --------------------
# ATTENUATION CURVES
# --------------------

__all__ = ["calzetti", "chevallard", "conroy", "noll",
           "powerlaw", "drude", "broken_powerlaw",
           "cardelli", "smc", "lmc"]


def powerlaw(wave, tau_v=1, alpha=1.0, **kwargs):
    """Simple power-law attenuation, normalized to 5500\AA.

    :param wave:
        The wavelengths at which optical depth estimates are desired.

    :param tau_v: (default: 1)
        The optical depth at 5500\AA, used to normalize the
        attenuation curve.

    :returns tau:
        The optical depth at each wavelength.
    """
    return tau_v * (wave / 5500)**(-alpha)


def calzetti(wave, tau_v=1, R_v=4.05, **kwargs):
    """Calzetti et al. 2000 starburst attenuation curve, with extrapolations to
    the FUV and NIR.

    :param wave:
        The wavelengths at which optical depth estimates are desired.

    :param tau_v: (default: 1)
        The optical depth at 5500\AA, used to normalize the attenuation curve.

    :param R_v: (default: 4.05)
        The ratio of total selective extinction, parameterizing the slope of
        the attenuation curve.  A_v = R_v * E(B-V)

    :returns tau:
        The optical depth at each wavelength.
    """
    # optical/NIR
    k1 = lambda x: 2.659 * (-1.857 + 1.040 * x)
    # UV
    k2 = lambda x: 2.659 * (-2.156 + 1.509 * x - 0.198 * x**2. + 0.011 * x**3.)

    # get slopes at edges and k(5500)
    uv = np.array([0.12, 0.13]) * 1e4
    kuv = k2(1e4 / uv) + R_v
    uv_slope = np.diff(kuv) / np.diff(uv)
    ir = np.array([2.19, 2.20]) * 1e4
    kir = k1(1e4 / ir) + R_v
    ir_slope = np.diff(kir) / np.diff(ir)
    k_v = k2(1e4 / 5500.) + R_v

    # define segments
    uinds = (wave >= 1200.) & (wave < 6300)  # uv
    oinds = (wave >= 6300.) & (wave <= 22000)  # optical
    xinds = (wave < 1200.)  # xuv
    iinds = (wave > 22000.)  # ir

    # do it
    x = 1e4 / wave
    ktot = oinds * (k1(x) + R_v)
    ktot += uinds * (k2(x) + R_v)
    ktot += xinds * (kuv[0] + (wave - uv[0]) * uv_slope)
    ktot += iinds * (kir[1] + (wave - ir[1]) * ir_slope)

    ktot[ktot < 0] = 0
    tau_lambda = tau_v * (ktot / k_v)
    return tau_lambda


def drude(x, x0=4.59, gamma=0.90, **extras):
    """Drude profile for the 2175AA bump.

    :param x:
        Inverse wavelength (inverse microns) at which values for the drude
        profile are requested.

    :param gamma:
        Width of the Drude profile (inverse microns).

    :param x0:
       Center of the Drude profile (inverse microns).

    :returns k_lambda:
       The value of the Drude profile at x, normalized such that the peak is 1.
     """
    #return (w * gamma)**2 / ((w**2 - w0**2)**2 + (w * gamma)**2)
    return (x*gamma)**2 / ((x**2 - x0**2)**2 + (x * gamma)**2)


def noll(wave, tau_v=1, delta=0.0, c_r=0.0, Ebump=0.0, **kwargs):
    """Noll 2009 attenuation curve. This is based on the Calzetti curve, with
    added variable bump (as Drude) and overall slope change.  Any extra
    keywords are passed to the Drude (e.g. x0, gamma, both in inverse microns).

    :param wave:
        The wavelengths at which optical depth estimates are desired.

    :param tau_v: (default: 1)
        The optical depth at 5500\AA, used to normalize the attenuation curve.

    :param Ebump: (default: 0.0)
        Stength of the 2175\AA bump.  Normalizes the Drude profile.

    :param delta: (default 0.)
        Slope of the power-law that modifies the Calzetti curve.

    :param c_r:
        Constant used to alter R_v=A_V/E(B-V) of the curve.  To maintain the
        Calzetti R_v=4.05, use c_r = -delta.  Note that even with c_r = -delta
        the calzetti curve will not be recovered unless delta=0

    :returns tau:
        The optical depth at each wavelength.
    """
    kcalz = calzetti(wave, tau_v=1.0, R_v=4.05) - 1
    k = kcalz + Ebump / 4.05 * drude(1e4 / wave, **kwargs)
    a = (k * (1 - 1.12 * c_r) + 1) * (wave / 5500.)**delta
    return a * tau_v


def chevallard(wave, tau_v=1, **kwargs):
    """ \tau_v dependent attenuation curves matched to disk RT models,
    as in Chevallard et al. 2013.  No UV bump (or indeed tests in the
    UV at all).

    :param wave:
        The wavelengths at which optical depth estimates are desired.

    :param tau_v: (default: 1)
        The optical depth at 5500\AA, used to normalize the
        attenuation curve.

    :returns tau:
        The optical depth at each wavelength.
    """
    # missing a UV bump
    alpha_v = 2.8 / (1 + np.sqrt(tau_v))  # +/- 25%
    bb = 0.3 - 0.05 * tau_v  # +/- 10%
    alpha = alpha_v + bb * (wave * 1e-4 - 0.55)
    tau_lambda = tau_v * (wave / 5500.0)**(-alpha)
    return tau_lambda


def conroy(wave, tau_v=1, R_v=3.1, f_bump=0.6, **kwargs):
    """ Conroy & Schiminovich 2010 dust attenuation curves including a
    decreased UV bump.

    :param wave:
        The wavelengths at which optical depth estimates are desired.

    :param tau_v: (default: 1)
        The optical depth at 5500\AA, used to normalize the
        attenuation curve.

    :param R_v: (default: 3.1)
        The ratio of total selective extinction, parameterizing the
        slope of the attenuation curve.  A_v = R_v * E(B-V)

    :param f_bump: (default: 0.6)
        The strength of the 2175\AA UV bump, as a fraction of the bump
        strength in Cardelli et al. extinction curve.

    :returns tau:
        The optical depth at each wavelength.
    """
    x = 1e4 / wave
    nx = x.shape[0]
    a = np.zeros_like(x)
    b = np.zeros_like(x)

    # IR 0.909 - 3.3 micron
    ir = (x >= 0.3) & (x < 1.1)
    a[ir] = 0.574 * x[ir]**1.61
    b[ir] = -0.527 * x[ir]**1.61

    # optical 0.303 - 0.909 micron
    opt = (x >= 1.1) & (x < 3.3)
    y = x[opt]-1.82
    a[opt] = (1 + 0.177 * y - 0.504 * y**2 - 0.0243 * y**3 +
              0.721 * y**4 + 0.0198 * y**5 - 0.7750 * y**6 +
              0.330 * y**7)
    b[opt] = (1.413 * y + 2.283 * y**2 + 1.072 * y**3 -
              5.384 * y**4 - 0.622 * y**5 + 5.303 * y**6 -
              2.090 * y**7)

    # NUV 0.17 to 0.303 micron
    nuv = (x >= 3.3) & (x < 5.9)
    tmp = (-0.0370 + 0.0469 * f_bump - 0.601 * f_bump / R_v + 0.542 / R_v)
    fa = (3.3 / x[nuv])**6. * tmp
    tmp = 0.104 * f_bump / ((x[nuv] - 4.67)**2 + 0.341)
    a[nuv] = 1.752 - 0.316 * x[nuv] - tmp + fa
    tmp = 1.206 * f_bump / ((x[nuv] - 4.62)**2 + 0.263)
    b[nuv] = -3.09 + 1.825 * x[nuv] + tmp

    # FUV 0.125 - 0.17 micron
    fuv = (x >= 5.9) & (x < 8.0)
    fa = -0.0447 * (x[fuv] - 5.9)**2.0 - 0.00978 * (x[fuv] - 5.9)**3
    fb = 0.213 * (x[fuv] - 5.9)**2. + 0.121 * (x[fuv] - 5.9)**3
    tmp = 0.104 * f_bump / ((x[fuv] - 4.67)**2 + 0.341)
    a[fuv] = 1.752 - 0.316 * x[fuv] - tmp + fa
    tmp = 1.206 * f_bump / ((x[fuv] - 4.62)**2 + 0.263)
    b[fuv] = -3.09 + 1.825 * x[fuv] + tmp + fb

    alam = (a + b / R_v)

    # XUV below 1250AA
    xuv = x >= 8.0
    x8 = 8.0
    fa = -0.0447 * (x8 - 5.9)**2 - 0.00978 * (x8 - 5.9)**3
    fb = 0.213 * (x8 - 5.9)**2. + 0.121 * (x8 - 5.9)**3
    tmp = 0.104 * f_bump / ((x8 - 4.67)**2 + 0.341)
    af = 1.752 - 0.316 * x8 - tmp + fa
    tmp = 1.206 * f_bump / ((x8 - 4.62)**2 + 0.263)
    bf = -3.09 + 1.825 * x8 + tmp + fb
    a8 = (af + bf / R_v)

    alam[xuv] = (x8 / x[xuv])**(-1.3) * a8

    return tau_v * alam


def broken_powerlaw(wave, tau_v=1, alpha=[0.7, 0.7, 0.7],
                    breaks=[0, 3000, 10000, 4e4], **kwargs):
    """ Attenuation curve as in V. Wild et al. 2011, i.e. power-law
    slope can change between regions.  Superceded by Chevallard 2013
    for optical/NIR.

    :param wave:
        The wavelengths at which optical depth estimates are desired.

    :param tau_v: (default: 1)
        The optical depth at 5500\AA, used to normalize the
        attenuation curve.

    :returns tau:
        The optical depth at each wavelength.
    """
    if len(breaks) == len(alpha)+1:
        print("make sure of your power law breaks")
    tau = np.array(len(wave))
    for i in range(alpha):
        inds = (wave > breaks[i]) & (wave <= breaks[i+1])
        tau[inds] = tau_v * (wave / 5500)**alpha[i]
    return tau


def wg00(wave, tau_v=1, geometry='SHELL', composition='MW',
         local='homogenous', **kwargs):
    """ Witt+Gordon 2000 DIRTY radiative transfer results, for
    idealized geometries.
    """
    raise NotImplementedError

# ------------------
# EXTINCTION CURVES
# ------------------


def cardelli(wave, tau_v=1, R_v=3.1, **kwargs):
    """ Cardelli, Clayton, and Mathis 1998 Milky Way extinction curve,
    with an update in the near-UV from O'Donnell 1994

    :param wave:
        The wavelengths at which optical depth estimates are desired.

    :param tau_v: (default: 1)
        The optical depth at 5500\AA, used to normalize the
        attenuation curve.

    :param R_v: (default: 3.1)
        The ratio of total selective extinction, parameterizing the
        slope of the attenuation curve.  A_v = R_v * E(B-V)

    :returns tau:
        The optical depth at each wavelength.
    """
    # if (wave < 1e3).any() :
    #     warnings.warn('Cardelli: extinction not defined (set to zero) below 1000AA')
    mic = wave*1e-4
    x_sup, x_inf = 10.0, 0.3
    x = 1 / mic
    a = np.zeros_like(x)
    b = np.zeros_like(x)

    w1 = (x >= 1.1) & (x <= 3.3)  # Optical 0.303 to 0.909 micron
    w2 = (x >= x_inf) & (x < 1.1)  # NIR 0.909 to 3.3 micron
    w3 = (x > 3.3) & (x <= 8)  # UV 0.125 - 0.303 micron
    w4 = (x > 8.0) & (x <= x_sup)  # XUV, 1000 -1250AA
    wsh = x > x_sup
    wlg = x < x_inf

    y = x[w1] - 1.82
    a[w1] = (1 + 0.17699 * y - 0.50447 * y**2. - 0.02427 * y**3. +
             0.72085 * y**4. + 0.01979 * y**5. - 0.77530 * y**6. +
             0.32999 * y**7.0)
    b[w1] = (1.41338 * y + 2.28305 * y**2. + 1.07233 * y**3. -
             5.38434 * y**4. - 0.62251 * y**5. + 5.30260 * y**6. -
             2.09002 * y**7.)

    y = x[w2]**1.61
    a[w2] = 0.574 * y
    b[w2] = -0.527 * y

    fa = x[w3] * 0.
    fb = x[w3] * 0.
    ou = (x[w3] > 5.9)
    # print(type(ou),ou[0], type(w3))

    if ou.any():
        y = x[w3][ou] - 5.9
        fa[ou] = -0.04473 * y**2. - 0.009779 * y**3.
        fb[ou] = 0.2130 * y**2. + 0.1207 * y**3.
    a[w3] = 1.752 - 0.316 * x[w3] - 0.104 / ((x[w3] - 4.67)**2. + 0.341) + fa
    b[w3] = -3.090 + 1.825 * x[w3] + 1.206 / ((x[w3] - 4.62)**2. + 0.263) + fb

    y = x[w4] - 8.
    a[w4] = -1.073 - 0.628 * y + 0.137 * y**2. - 0.070 * y**3.
    b[w4] = 13.670 + 4.257 * y - 0.420 * y**2. + 0.374 * y**3.

    tau = a + b / R_v
    return tau_v * tau


def smc(wave, tau_v=1, **kwargs):
    """Pei 1992 SMC extinction curve.

    :param wave:
        The wavelengths at which optical depth estimates are desired.

    :param tau_v: (default: 1)
        The optical depth at 5500\AA, used to normalize the
        attenuation curve.

    :returns tau:
        The optical depth at each wavelength.
    """
    if (wave < 1e3).any():
        warnings.warn('SMC: extinction extrapolation below 1000AA is poor')
    mic = wave * 1e-4
    aa = [185.,  27.,  0.005, 0.010, 0.012, 0.030]
    ll = [0.042, 0.08, 0.22,  9.7,   18.,   25.]
    bb = [90.,   5.50, -1.95, -1.95, -1.80, 0.00]
    nn = [2.0,   4.0,  2.0,   2.0,   2.0,   2.0]

    abs_ab = np.zeros_like(mic)
    norm_v = 0  # hack to go from tau_b to tau_v
    mic_5500 = 5500 * 1e-4

    for i, a in enumerate(aa):
        norm_v += aa[i] / ((mic_5500 / ll[i])**nn[i] +
                           (ll[i] / mic_5500)**nn[i] + bb[i])
        abs_ab += aa[i] / ((mic / ll[i])**nn[i] + (ll[i] / mic)**nn[i] + bb[i])

    return tau_v * (abs_ab / norm_v)


def lmc(wave, tau_v=1, **kwargs):
    """ Pei 1992 LMC extinction curve.

    :param wave:
        The wavelengths at which optical depth estimates are desired.

    :param tau_v: (default: 1)
        The optical depth at 5500\AA, used to normalize the
        attenuation curve.

    :returns tau:
        The optical depth at each wavelength.
    """
    if (wave < 1e3).any():
        warnings.warn('LMC: extinction extrapolation below 1000AA is poor')
    mic = wave * 1e-4
    aa = [175.,  19.,  0.023, 0.005, 0.006, 0.020]
    ll = [0.046, 0.08, 0.22,  9.7,   18.,   25.]
    bb = [90.,   5.50, -1.95, -1.95, -1.80, 0.00]
    nn = [2.0,   4.5,  2.0,   2.0,   2.0,   2.0]

    abs_ab = mic * 0.
    norm_v = 0  # hack to go from tau_b to tau_v
    mic_5500 = 5500 * 1e-4

    for i, a in enumerate(aa):
        norm_v += aa[i] / ((mic_5500 / ll[i])**nn[i] +
                           (ll[i] / mic_5500)**nn[i] + bb[i])
        abs_ab += aa[i] / ((mic / ll[i])**nn[i] + (ll[i] / mic)**nn[i] + bb[i])

    return tau_v * (abs_ab / norm_v)



# This code is an unfinished attempt to plot SEDs as probability
# distributions p(flux, wave)

import numpy as np

class Bunch(object):
    """ Simple storage.
    """
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class SedPoint(object):

    def __init__(self, flux=0., unc=0., band=None):

        self.flux = flux
        self.unc = unc
        self.band = band

    @property
    def var(self):
        return self.unc**2

    @property
    def pnorm(self):
        return (self.var * 2 * np.pi)**(-0.5)

    @property
    def wave_min(self):
        return self.band.wave_effective - self.band.effective_width/2.

    @property
    def wave_max(self):
        return self.band.wave_effective + self.band.effective_width/2.

    def flux_lnprob(self, fluxes):
        fnp =  (np.array(fluxes) - self.flux)**2 / (2 * self.var) 
        return  np.log(self.pnorm) - fnp

    def spans(self, wavelengths):
        waves = np.array(wavelengths)
        return (waves < self.wave_max) & (waves > self.wave_min)

    def sed_prob(self, wavelength_grid, flux_grid):
        return self.spans(wavelength_grid), self.flux_lnprob(flux_grid)


def sed_to_psed(filters, fluxes, uncs, wgrid, fgrid):
    sed = []
    lnpsed = np.zeros(( len(wgrid), len(fgrid)))
    for filt, flux, unc in zip(filters, fluxes, uncs):
        sed_point =  SedPoint(flux, unc, filt)
        winds, lnprob = sed_point.sed_prob(wgrid, fgrid)
        lnpsed[winds,:] += lnprob[None, :]
        sed += [sed_point]
        lnpsed -= np.log(np.trapz(np.exp(lnpsed), fgrid, axis = 1))[:, None]
        lnpsed += np.log(np.trapz(np.ones(len(fgrid)), fgrid))
    return lnpsed, sed


def test():
    from sedpy import observate
    import fsps
    import matplotlib.pyplot as pl

    filters = ['galex_NUV', 'sdss_u0', 'sdss_r0', 'sdss_r0', 'sdss_i0', 'sdss_z0',
               'bessell_U', 'bessell_B', 'bessell_V', 'bessell_R', 'bessell_I',
               'twomass_J','twomass_H']
    flist = observate.load_filters(filters)

    sps = fsps.StellarPopulation(compute_vega_mags=False)
    wave, spec = sps.get_spectrum(tage=1.0, zmet=2, peraa=True)

    sed = observate.getSED(wave, spec, flist)
    sed_unc = np.abs(np.random.normal(1, 0.3, len(sed)))

    wgrid = np.linspace( 2e3, 13e3, 1000)
    fgrid = np.linspace( -13, -9, 100)
    psed, sedpoints = sed_to_psed(flist, sed, sed_unc, wgrid, fgrid)

    pl.imshow(np.exp(psed).T, cmap='Greys_r',
              interpolation='nearest', origin ='upper', aspect='auto')
    



"""smoothing.py  - Spectral smoothing functionality
"""
# TODO:
#  * add extra zero-padding for FFT algorithms so they don't go funky at the
#    edges?
#  * more testing of the smooth_lsf* methods.

import numpy as np
from numpy.fft import fft, ifft, fftfreq, rfftfreq

__all__ = ["smoothspec", "smooth_wave", "smooth_vel", "smooth_lsf",
           "smooth_wave_fft", "smooth_vel_fft", "smooth_fft", "smooth_lsf_fft",
           "mask_wave", "resample_wave"]

ckms = 2.998e5
sigma_to_fwhm = 2.355


def smoothspec(wave, spec, resolution=None, outwave=None,
               smoothtype="vel", fftsmooth=True,
               min_wave_smooth=0, max_wave_smooth=np.inf, **kwargs):
    """
    Parameters
    ----------
    wave : ndarray of shape ``(N_pix,)``
        The wavelength vector of the input spectrum.  Assumed Angstroms.

    spec : ndarray of shape ``(N_pix,)``
        The flux vector of the input spectrum.

    resolution : float
        The smoothing parameter.  Units depend on ``smoothtype``.

    outwave : ``None`` or ndarray of shape ``(N_pix_out,)``
        The output wavelength vector.  If ``None`` then the input wavelength
        vector will be assumed, though if ``min_wave_smooth`` or
        ``max_wave_smooth`` are also specified, then the output spectrum may
        have different length than ``spec`` or ``wave``, or the convolution may
        be strange outside of ``min_wave_smooth`` and ``max_wave_smooth``.
        Basically, always set ``outwave`` to be safe.

    smoothtype : string, optional, default: "vel"
        The type of smoothing to perform.  One of:

        + ``"vel"`` - velocity smoothing, ``resolution`` units are in km/s
          (dispersion not FWHM)
        + ``"R"`` - resolution smoothing, ``resolution`` is in units of
          :math:`\lambda/ \sigma_\lambda` (where :math:`\sigma_\lambda` is
          dispersion, not FWHM)
        + ``"lambda"`` - wavelength smoothing.  ``resolution`` is in units of
          Angstroms
        + ``"lsf"`` - line-spread function.  Use an aribitrary line spread
          function, which can be given as a vector the same length as ``wave``
          that gives the dispersion (in AA) at each wavelength.  Alternatively,
          if ``resolution`` is ``None`` then a line-spread function must be
          present as an additional ``lsf`` keyword.  In this case all additional
          keywords as well as the ``wave`` vector will be passed to this ``lsf``
          function.

    fftsmooth : bool, optional, default: True
        Switch to use FFTs to do the smoothing, usually resulting in massive
        speedups of all algorithms.  However, edge effects may be present.

    min_wave_smooth : float, optional default: 0
        The minimum wavelength of the input vector to consider when smoothing
        the spectrum.  If ``None`` then it is determined from the output
        wavelength vector and padded by some multiple of the desired resolution.

    max_wave_smooth : float, optional, default: inf
        The maximum wavelength of the input vector to consider when smoothing
        the spectrum.  If None then it is determined from the output wavelength
        vector and padded by some multiple of the desired resolution.

    inres : float, optional
        If given, this parameter specifies the resolution of the input.  This
        resolution is subtracted in quadrature from the target output resolution
        before the kernel is formed.

        In certain cases this can be used to properly switch from resolution
        that is constant in velocity to one that is constant in wavelength,
        taking into account the wavelength dependence of the input resolution
        when defined in terms of lambda.  This is possible iff:
        * ``fftsmooth`` is False
        * ``smoothtype`` is ``"lambda"``
        * The optional ``in_vel`` parameter is supplied and True.

        The units of ``inres`` should be the same as the units of
        ``resolution``, except in the case of switching from velocity to
        wavelength resolution, in which case the units of ``inres`` should be
        in units of lambda/sigma_lambda.

    in_vel : float (optional)
        If supplied and True, the ``inres`` parameter is assumed to be in units
        of lambda/sigma_lambda. This parameter is ignored **unless** the
        ``smoothtype`` is ``"lambda"`` and ``fftsmooth`` is False.

    Returns
    -------
    flux : ndarray of shape ``(N_pix_out,)``
        The smoothed spectrum on the `outwave` grid, ndarray.
    """
    if smoothtype == 'vel':
        linear = False
        units = 'km/s'
        sigma = resolution
        fwhm = sigma * sigma_to_fwhm
        Rsigma = ckms / sigma
        R = ckms / fwhm
        width = Rsigma
        assert np.size(sigma) == 1, "`resolution` must be scalar for `smoothtype`='vel'"

    elif smoothtype == 'R':
        linear = False
        units = 'km/s'
        Rsigma = resolution
        sigma = ckms / Rsigma
        fwhm = sigma * sigma_to_fwhm
        R = ckms / fwhm
        width = Rsigma
        assert np.size(sigma) == 1, "`resolution` must be scalar for `smoothtype`='R'"
        # convert inres from Rsigma to sigma (km/s)
        try:
            kwargs['inres'] = ckms / kwargs['inres']
        except(KeyError):
            pass

    elif smoothtype == 'lambda':
        linear = True
        units = 'AA'
        sigma = resolution
        fwhm = sigma * sigma_to_fwhm
        Rsigma = None
        R = None
        width = sigma
        assert np.size(sigma) == 1, "`resolution` must be scalar for `smoothtype`='lambda'"

    elif smoothtype == 'lsf':
        linear = True
        width = 100
        sigma = resolution

    else:
        raise ValueError("smoothtype {} is not valid".format(smoothtype))

    # Mask the input spectrum depending on outwave or the wave_smooth kwargs
    mask = mask_wave(wave, width=width, outwave=outwave, linear=linear,
                     wlo=min_wave_smooth, whi=max_wave_smooth, **kwargs)
    w = wave[mask]
    s = spec[mask]
    if outwave is None:
        outwave = wave

    # Choose the smoothing method
    if smoothtype == 'lsf':
        if fftsmooth:
            smooth_method = smooth_lsf_fft
            if sigma is not None:
                # mask the resolution vector
                sigma = resolution[mask]
        else:
            smooth_method = smooth_lsf
            if sigma is not None:
                # convert to resolution on the output wavelength grid
                sigma = np.interp(outwave, wave, resolution)
    elif linear:
        if fftsmooth:
            smooth_method = smooth_wave_fft
        else:
            smooth_method = smooth_wave
    else:
        if fftsmooth:
            smooth_method = smooth_vel_fft
        else:
            smooth_method = smooth_vel

    # Actually do the smoothing and return
    return smooth_method(w, s, outwave, sigma, **kwargs)


def smooth_vel(wave, spec, outwave, sigma, nsigma=10, inres=0, **extras):
    """Smooth a spectrum in velocity space.  This is insanely slow, but general
    and correct.

    Parameters
    ----------
    wave : ndarray of shape ``(N_pix,)``
        Wavelength vector of the input spectrum.

    spec : ndarray of shape ``(N_pix,)``
        Flux vector of the input spectrum.

    outwave : ndarray of shape ``(N_pix_out,)``
        Desired output wavelength vector.

    sigma : float
        Desired velocity resolution (km/s), dispersion *not* FWHM.

    nsigma : float (optional)
        Number of sigma away from the output wavelength to consider in the
        integral.  If less than zero, all wavelengths are used.  Setting this
        to some positive number decreses the scaling constant in the O(N_out *
        N_in) algorithm used here.

    inres : float (optional)
        The velocity resolution of the input spectrum (km/s), *not* FWHM. the
        spectrum will be smoothed by the quadrature difference between the
        desired resolution and ``inres``

    Returns
    -------
    smoothed_spec : ndarray of shape ``(N_pix_out,)``
        The smoothed spectrum
    """
    sigma_eff_sq = sigma**2 - inres**2
    if np.any(sigma_eff_sq) < 0.0:
        raise ValueError("Desired velocity resolution smaller than the value"
                         "possible for this input spectrum.".format(inres))
    # sigma_eff is in units of sigma_lambda / lambda
    sigma_eff = np.sqrt(sigma_eff_sq) / ckms

    lnwave = np.log(wave)
    flux = np.zeros(len(outwave))
    for i, w in enumerate(outwave):
        x = (np.log(w) - lnwave) / sigma_eff
        if nsigma > 0:
            good = np.abs(x) < nsigma
            x = x[good]
            _spec = spec[good]
        else:
            _spec = spec
        f = np.exp(-0.5 * x**2)
        flux[i] = np.trapz(f * _spec, x) / np.trapz(f, x)
    return flux


def smooth_vel_fft(wavelength, spectrum, outwave, sigma_out, inres=0.0,
                   **extras):
    """Smooth a spectrum in velocity space, using FFTs. This is fast, but makes
    some assumptions about the form of the input spectrum and can have some
    issues at the ends of the spectrum depending on how it is padded.

    Parameters
    ----------
    wavelength : ndarray of shape ``(N_pix,)``
        Wavelength vector of the input spectrum.

    spectrum : ndarray of shape ``(N_pix,)``
        Flux vector of the input spectrum.

    outwave : ndarray of shape ``(N_pix_out,)``
        Desired output wavelength vector.

    sigma_out : float
        Desired velocity resolution (km/s), dispersion *not* FWHM

    inres : float (optional)
        The velocity resolution of the input spectrum (km/s), *not* FWHM. the
        spectrum will be smoothed by the quadrature difference between the
        desired resolution and ``inres``

    Returns
    -------
    smoothed_spec : ndarray of shape ``(N_pix_out,)``
        The smoothed spectrum
    """
    # The kernel width for the convolution.
    sigma = np.sqrt(sigma_out**2 - inres**2)
    if sigma <= 0:
        return np.interp(outwave, wavelength, spectrum)

    # make length of spectrum a power of 2 by resampling
    wave, spec = resample_wave(wavelength, spectrum)

    # get grid resolution (*not* the resolution of the input spectrum) and make
    # sure it's nearly constant.  It should be, by design (see resample_wave)
    invRgrid = np.diff(np.log(wave))
    assert invRgrid.max() / invRgrid.min() < 1.05
    dv = ckms * np.median(invRgrid)

    # Do the convolution
    spec_conv = smooth_fft(dv, spec, sigma)
    # interpolate onto output grid
    if outwave is not None:
        spec_conv = np.interp(outwave, wave, spec_conv)

    return spec_conv


def smooth_wave(wave, spec, outwave, sigma, nsigma=10, inres=0, in_vel=False,
                **extras):
    """Smooth a spectrum in wavelength space.  This is insanely slow, but
    general and correct (except for the treatment of the input resolution if it
    is velocity)

    Parameters
    ----------
    wave : ndarray of shape ``(N_pix,)``
        Wavelength vector of the input spectrum.

    spec : ndarray of shape ``(N_pix,)``
        Flux vector of the input spectrum.

    outwave : ndarray of shape ``(N_pix_out,)``
        Desired output wavelength vector.

    sigma : float or ndarray of shape ``(N_pix,)``
        Desired resolution (dispersion *not* FWHM) in wavelength units.  This
        can be a vector of same length as ``wave``, in which case a wavelength
        dependent broadening is calculated

    nsigma : float (optional, default=10)
        Number of sigma away from the output wavelength to consider in the
        integral.  If less than zero, all wavelengths are used.  Setting this to
        some positive number decreses the scaling constant in the O(N_out *
        N_in) algorithm used here.

    inres : float (optional)
        Resolution of the input, in either wavelength units or
        :math:`\lambda/d\lambda \cdot (c/v)`. The spectrum will be smoothed by
        the quadrature difference between the desired resolution and ``inres``

    in_vel : bool (optional, default: False)
        If True, the input spectrum has been smoothed in velocity
        space, and ``inres`` is assumed to be in lambda/dlambda.

    Returns
    -------
    smoothed_spec : ndarray of shape ``(N_pix_out,)``
        The smoothed spectrum.
    """
    # sigma_eff is in angstroms
    if inres <= 0:
        sigma_eff_sq = sigma**2
    elif in_vel:
        # Make an approximate correction for the intrinsic wavelength
        # dependent dispersion.  This sort of maybe works.
        sigma_eff_sq = sigma**2 - (wave / inres)**2
    else:
        sigma_eff_sq = sigma**2 - inres**2
    if np.any(sigma_eff_sq < 0):
        raise ValueError("Desired wavelength sigma is lower than the value "
                         "possible for this input spectrum.")

    sigma_eff = np.sqrt(sigma_eff_sq)
    flux = np.zeros(len(outwave))
    for i, w in enumerate(outwave):
        x = (wave - w) / sigma_eff
        if nsigma > 0:
            good = np.abs(x) < nsigma
            x = x[good]
            _spec = spec[good]
        else:
            _spec = spec
        f = np.exp(-0.5 * x**2)
        flux[i] = np.trapz(f * _spec, x) / np.trapz(f, x)
    return flux


def smooth_wave_fft(wavelength, spectrum, outwave, sigma_out=1.0,
                    inres=0.0, **extras):
    """Smooth a spectrum in wavelength space, using FFTs.  This is fast, but
    makes some assumptions about the input spectrum, and can have some
    issues at the ends of the spectrum depending on how it is padded.

    Parameters
    ----------
    wavelength : ndarray of shape ``(N_pix,)``
        Wavelength vector of the input spectrum.

    spectrum : ndarray of shape ``(N_pix,)``
        Flux vector of the input spectrum.

    outwave : ndarray of shape ``(N_pix_out,)``
        Desired output wavelength vector.

    sigma : float
        Desired resolution (dispersion *not* FWHM) in wavelength units.

    inres : float, optional
        Resolution of the input, in wavelength units (dispersion not FWHM). The
        spectrum will be smoothed by the quadrature difference between the
        desired resolution and ``inres``

    Returns
    -------
    smoothed_spec : ndarray of shape ``(N_pix_out,)``
        The smoothed spectrum.
    """
    # restrict wavelength range (for speed)
    # should also make nearest power of 2
    wave, spec = resample_wave(wavelength, spectrum, linear=True)

    # The kernel width for the convolution.
    sigma = np.sqrt(sigma_out**2 - inres**2)
    if sigma < 0:
        return np.interp(wave, outwave, flux)

    # get grid resolution (*not* the resolution of the input spectrum) and make
    # sure it's nearly constant.  Should be by design (see resample_wave)
    Rgrid = np.diff(wave)
    assert Rgrid.max() / Rgrid.min() < 1.05
    dw = np.median(Rgrid)

    # Do the convolution
    spec_conv = smooth_fft(dw, spec, sigma)
    # interpolate onto output grid
    if outwave is not None:
        spec_conv = np.interp(outwave, wave, spec_conv)
    return spec_conv


def smooth_lsf(wave, spec, outwave, sigma=None, lsf=None, return_kernel=False,
               **lsf_kwargs):
    """Broaden a spectrum using a wavelength dependent line spread function.
    This function is only approximate because it doesn't actually do the
    integration over pixels, so for sparsely sampled points you'll have
    problems.  This function needs to be checked and possibly rewritten.

    Parameters
    ----------
    wavelength : ndarray of shape ``(N_pix,)``
        Wavelength vector of the input spectrum.

    spectrum : ndarray of shape ``(N_pix,)``
        Flux vector of the input spectrum.

    outwave : ndarray of shape ``(N_pix_out,)``
        Desired output wavelength vector.

    sigma : float (optional, default: None)
        The dispersion (not FWHM) as a function of wavelength that you want to
        apply to the input spectrum.  ``None`` or ndarray of same length as
        ``outwave``.  If ``None`` then the wavelength dependent dispersion will
        be calculated from the function supplied with the ``lsf`` keyward.

    lsf : callable (optional)
        A function that returns the Gaussian dispersion at each wavelength.
        This is assumed to be in sigma, not FWHM.

    return_kernel : bool (optional, default: False)
        If True, return the kernel used to broaden the spectrum as ndarray of
        shape ``(nout, nin)``.

    Extra Parameters
    ----------------
    lsf_kwargs : dict (optional)
        Passed to the function supplied in the ``lsf`` keyword.

    Returns
    -------
    smoothed_spec : ndarray of shape ``(N_pix_out,)``
        The smoothed spectrum.
    """
    if (lsf is None) and (sigma is None):
        return np.interp(outwave, wave, spec)
    dw = np.gradient(wave)
    if sigma is None:
        sigma = lsf(outwave, **lsf_kwargs)
    kernel = outwave[:, None] - wave[None, :]
    kernel = (1 / (sigma * np.sqrt(np.pi * 2))[:, None] *
              np.exp(-kernel**2 / (2 * sigma[:, None]**2)) *
              dw[None, :])
    # should this be axis=0 or axis=1?
    kernel = kernel / kernel.sum(axis=1)[:, None]
    newspec = np.dot(kernel, spec)
    # kernel /= np.trapz(kernel, wave, axis=1)[:, None]
    # newspec = np.trapz(kernel * spec[None, :], wave, axis=1)
    if return_kernel:
        return newspec, kernel
    return newspec


def smooth_lsf_fft(wave, spec, outwave, sigma=None, lsf=None, pix_per_sigma=2,
                   eps=0.25, preserve_all_input_frequencies=False, **kwargs):
    """Smooth a spectrum by a wavelength dependent line-spread function, using
    FFTs.

    Parameters
    ----------
    wavelength : ndarray of shape ``(N_pix,)``
        Wavelength vector of the input spectrum.

    spectrum : ndarray of shape ``(N_pix,)``
        Flux vector of the input spectrum.

    outwave : ndarray of shape ``(N_pix_out,)``
        Desired output wavelength vector.

    sigma : ndarray of shape ``(N_pix,)`` (optional)
        Dispersion (in same units as ``wave``) as a function `wave`.  If not
        given, sigma will be computed from the function provided by the ``lsf``
        keyword.

    lsf : callable (optional)
        Function used to calculate the dispersion as a function of wavelength.
        Must be able to take as an argument the ``wave`` vector and any extra
        keyword arguments and return the dispersion (in the same units as the
        input wavelength vector) at every value of ``wave``.  If not provided
        then ``sigma`` must be specified.

    pix_per_sigma : float (optional, default: 2)
        Number of pixels per sigma of the smoothed spectrum to use in
        intermediate interpolation and FFT steps. Increasing this number will
        increase the accuracy of the output (to a point), and the run-time, by
        preserving all high-frequency information in the input spectrum.

    preserve_all_input_frequencies : bool (default: False)
        This is a switch to use a very dense sampling of the input spectrum that
        preserves all input frequencies.  It can significantly increase the call
        time for often modest gains...

    eps : float (optional)
        Deprecated.

    Extra Parameters
    ----------------
    kwargs:
        All additional keywords are passed to the function supplied to the
        ``lsf`` keyword, if present.

    Returns
    -------
    smoothed_spec : ndarray of shape ``(N_pix_out,)``
        The smoothed spectrum.
    """
    # This is sigma vs lambda
    if sigma is None:
        sigma = lsf(wave, **kwargs)

    # Now we need the CDF of 1/sigma, which provides the relationship between x and lambda
    # does dw go in numerator or denominator?
    # I think numerator but should be tested
    dw = np.gradient(wave)
    cdf = np.cumsum(dw / sigma)
    cdf /= cdf.max()

    # Now we create an evenly sampled grid in the x coordinate on the interval [0,1]
    # and convert that to lambda using the cdf.
    # This should result in some power of two x points, for FFT efficiency

    # Furthermore, the number of points should be high enough that the
    # resolution is critically sampled.  And we want to know what the
    # resolution is in this new coordinate.
    # There are two possible ways to do this

    # 1) Choose a point ~halfway in the spectrum
    # half = len(wave) / 2
    # Now get the x coordinates of a point eps*sigma redder and bluer
    # wave_eps = eps * np.array([-1, 1]) * sigma[halpha]
    # x_h_eps = np.interp(wave[half] + wave_eps, wave, cdf)
    # Take the differences to get dx and dsigma and ratio to get x per sigma
    # x_per_sigma = np.diff(x_h_eps) / (2.0 * eps) #x_h_epsilon - x_h

    # 2) Get for all points (slower?):
    sigma_per_pixel = (dw / sigma)
    x_per_pixel = np.gradient(cdf)
    x_per_sigma = np.nanmedian(x_per_pixel / sigma_per_pixel)
    N = pix_per_sigma / x_per_sigma

    # Alternatively, just use the smallest dx of the input, divided by two for safety
    # Assumes the input spectrum is critically sampled.
    # And does not actually give x_per_sigma, so that has to be determined anyway
    if preserve_all_input_frequencies:
        # preserve more information in the input spectrum, even when way higher
        # frequency than the resolution of the output.  Leads to slightly more
        # accurate output, but with a substantial time hit
        N = max(N, 1.0 / np.nanmin(x_per_pixel))

    # Now find the smallest power of two that divides the interval (0, 1) into
    # segments that are smaller than dx
    nx = int(2**np.ceil(np.log2(N)))

    # now evenly sample in the x coordinate
    x = np.linspace(0, 1, nx)
    dx = 1.0 / nx

    # And now we get the spectrum at the lambda coordinates of the even grid in x
    lam = np.interp(x, cdf, wave)
    newspec = np.interp(lam, wave, spec)

    # And now we convolve.
    # If we did not know sigma in terms of x we could estimate it here
    # from the resulting sigma(lamda(x)) / dlambda(x):
    # dlam = np.gradient(lam)
    # sigma_x = np.median(lsf(lam, **kwargs) / dlam)
    # But the following just uses the fact that we know x_per_sigma (duh).
    spec_conv = smooth_fft(dx, newspec, x_per_sigma)

    # and interpolate back to the output wavelength grid.
    return np.interp(outwave, lam, spec_conv)


def smooth_fft(dx, spec, sigma):
    """Basic math for FFT convolution with a gaussian kernel.

    Parameters
    ----------
    dx : ndarray of shape ``(N_pix,)``
        The wavelength or velocity spacing, same units as ``sigma``

    spec : ndarry of shape ``(N_pix,)``
        The spectrum flux vector

    sigma : float
        The width of the Gaussian kernel, same units as ``dx``

    Returns
    -------
    smoothed_spec : ndarry of shape ``(N_pix,)``
        The smoothed spectrum, on the same grid as the input.
    """
    # The Fourier coordinate
    ss = rfftfreq(len(spec), d=dx)
    # Make the fourier space taper; just the analytical fft of a gaussian
    taper = np.exp(-2 * (np.pi ** 2) * (sigma ** 2) * (ss ** 2))
    ss[0] = 0.01  # hack
    # Fourier transform the spectrum
    spec_ff = np.fft.rfft(spec)
    # Multiply in fourier space
    ff_tapered = spec_ff * taper
    # Fourier transform back
    spec_conv = np.fft.irfft(ff_tapered)
    return spec_conv


def mask_wave(wavelength, width=1, wlo=0, whi=np.inf, outwave=None,
              nsigma_pad=20.0, linear=False, **extras):
    """Restrict wavelength range (for speed) but include some padding based on
    the desired resolution.
    """
    # Base wavelength limits
    if outwave is not None:
        wlim = np.array([outwave.min(), outwave.max()])
    else:
        wlim = np.squeeze(np.array([wlo, whi]))
    # Pad by nsigma * sigma_wave
    if linear:
        wlim += nsigma_pad * width * np.array([-1, 1])
    else:
        wlim *= (1 + nsigma_pad / width * np.array([-1, 1]))
    mask = (wavelength > wlim[0]) & (wavelength < wlim[1])
    return mask


def resample_wave(wavelength, spectrum, linear=False):
    """Resample spectrum, so that the number of elements is the next highest
    power of two.  This uses np.interp.  Note that if the input wavelength grid
    did not critically sample the spectrum then there is no gaurantee the
    output wavelength grid will.
    """
    wmin, wmax = wavelength.min(), wavelength.max()
    nw = len(wavelength)
    nnew = int(2.0**(np.ceil(np.log2(nw))))
    if linear:
        Rgrid = np.diff(wavelength)  # in same units as ``wavelength``
        w = np.linspace(wmin, wmax, nnew)
    else:
        Rgrid = np.diff(np.log(wavelength))  # actually 1/R
        lnlam = np.linspace(np.log(wmin), np.log(wmax), nnew)
        w = np.exp(lnlam)
    # Make sure the resolution really is nearly constant
    #assert Rgrid.max() / Rgrid.min() < 1.05
    s = np.interp(w, wavelength, spectrum)
    return w, s


def subtract_input_resolution(res_in, res_target, smoothtype_in, smoothtype_target, wave=None):
    """Subtract the input resolution (in quadrature) from a target output
    resolution to get the width of the kernel that will convolve the input to
    the output.  Assumes all convolutions are with gaussians.
    """
    if smoothtype_in == "R":
        width_in = 1.0 / res_in
    else:
        width_in = res_in
    if smoothtype_target == "R":
        width_target = 1.0 / res_target
    else:
        width_target = res_target

    if smoothtype_in == smoothtype_target:
        dwidth_sq = width_target**2 - width_in**2

    elif (smoothtype_in == "vel") & (smoothype_target == "lambda"):
        dwidth_sq = width_target**2 - (wave * width_in / ckms)**2

    elif (smoothtype_in == "R") & (smoothype_target == "lambda"):
        dwidth_sq = width_target**2 - (wave * width_in)**2

    elif (smoothtype_in == "lambda") & (smoothtype_target == "vel"):
        dwidth_sq = width_target**2 - (ckms * width_in / wave)**2

    elif (smoothtype_in == "lambda") & (smoothtype_target == "R"):
        dwidth_sq = width_target**2 - (width_in / wave)**2

    elif (smoothtype_in == "R") & (smoothtype_target == "vel"):
        print("srsly?")
        return None
    elif (smoothtype_in == "vel") & (smoothtype_target == "R"):
        print("srsly?")
        return None

    if np.any(dwidth_sq <= 0):
        print("Warning:  Desired resolution is better than input resolution")
        dwidth_sq = np.clip(dwidth_sq, 0, np.inf)

    if smoothtype_target == "R":
        return 1.0 / np.sqrt(dwidth_sq)
    else:
        return np.sqrt(dwidth_sq)

    return delta_width



# -*- coding: utf-8 -*-

"""extinction.py  - Classes for implementing detailed dust extinction models.

Old code, use with caution!
"""

import numpy as np
import warnings, sys
from numpy.random import normal
try:
    import scipy.interpolate as interp
except:
    pass

thismod = sys.modules[__name__]


class Attenuator(object):
    """
    A level of abstraction around the dust curves, which allows for
    some generic methods. The Attenuator is composed of an (effective)
    DustLaw and a DustDistribution class, both of which can be easily
    extended.  The first takes wavelength and tau_v as arguments and
    returns \tau_\lambda. The second takes arguments that can be
    anything (e.g. stellar age, metallicity) and returns a
    distribution of tau_V
    """

    def __init__(self, dust_type='FM07'):
        self.name = dust_type
        self.dustcurve = getattr(thismod, dust_type)()
        # self.dustdist =

    def attenuate_spectrum(self, wave, inspec, pars):
        """This will be slow until the dust curves are vectorized.
        """
        spec = np.atleast_2d(inspec)  # memory hog
        for i, p in enumerate(pars):
            tau = self.dustcurve.acurve(wave, A_v=p['A_V'],
                                        R_v=p['R_V'],
                                        f_bump=p['F_BUMP'],
                                        uv_slope=p['UV_SLOPE'])
            spec[i, :] = spec[i, :]*np.exp(-tau)
        return spec

    def draw_taus(self, ntau, pars):
        """Not implemented
        """
        pass

# ------------------------------
# Extinction Law Classes
# ------------------------------


class GenericCurve(object):
    """Class for handling extinction curves with the parameterization
    of Fitzpatrick & Massa 1990 or F&M 2007
    """

    def __init__(self, form='F07', var=False):
        self.default_pars(var=var)

        if form == 'FM90':
            self._f = self.f_90
        elif form == 'F07':
            self._f = self.f_07

    def fm_curve(self, x, c1=None, c2=None, c3=None, c4=None,
                 c5=None, x0=4.59, gamma=0.89, **extras):
        """Fitzpatrick & Massa parameterization of the UV excess
        curve.
        """
        d = self.drude(x, gamma, x0)
        f = self._f(x, c5=c5)
        return (c1 + c2 * x + c3 * d + c4 * f)

    def drude(self, x, gamma, x0):
        """Drude profile for the 2175AA bump.

        :param x:
           Inverse wavelength (inverse microns) at which
           values for the drude profile are requested.

        :param gamma:
           Width of the Drude profile (inverse microns).

        :param x0:
           Center of the Drude profile (inverse microns).

        :returns k_lambda:
           The value of the Drude profile at x
        """
        return x**2 / ((x**2 - x0**2)**2 + (x * gamma)**2)

    def f_90(self, x, **extras):
        """UV rise in the excess curve, from FM90 (defined cubic
        polynomial).

        :param x:
            Inverse wavelength in inverse microns

        :returns:
            Cubic polynomial at x centered on 5.9 microns**(-1), as in
            FM90
        """
        return (x > 5.9)*(0.5392*(x-5.9)*(x-5.9)+0.05644*(x-5.9)**3)

    def f_07(self, x, c5=5.9, **extras):
        """UV rise in the excess curve, from FM07 (quadratic with
        variable center).

        :param x:
            Inverse wavelength in inverse microns.

        :param c5:
            Pivot point for the quadratic.

        :returns:
            Quadratic at x centered on c5, as in FM07.
        """
        return (x > c5)*(x-c5)*(x-c5)

    def powerlaw(self, x, R_v=3.1, k=None, alpha=-1.84, **extras):
        """Power-law shape of the color excess curve, for the NIR
        extinction curve, as in FM09.

        :param x:
            Inverse wavelength in inverse microns.

        :param R_v: (default: 3.1)
            Slope of the optical portion of the curve.

        :param k: (default: None)
            Power-law normalization.

        :param alpha: (default: -1.84)
            Power-law slope.
        """
        return k*(1./x)**alpha - R_v

    def spline(self, x, spline_x=[1.8, 2.5, 3.0],
               spline_k=[0.0, 1.32, 2.02]):
        """Cubic spline, for the optical portion of the extinction
        curve.

        :param x:
            Inverse wavelength in inverse microns.

        :param spline_x:
            Location of the spline anchor points, in inverse microns.

        :param spline_k:
            Value at the spline anchor points.
        """
        spline_x = np.asarray(spline_x)
        spline_k = np.asarray(spline_k)
        sp = interp.InterpolatedUnivariateSpline(spline_x, spline_k)
        return sp(x)

    def acurve(self, wave, **pars):
        """Return the extinction curve given the excess curve.
        """
        x = 1e4/wave
        self.setpars(**pars)
        e = self.ecurve(x, self.pardict)
        return e * self.pardict['A_v']/self.pardict['R_v'] + 1

    def setpars(self, A_v=1, R_v=3.1, f_bump=1.0, uv_slope=1,
                var=False, **extras):
        self.default_pars(var=var)
        self.pardict['c3'] *= f_bump
        self.pardict['A_v'] = A_v
        self.pardict['R_v'] = R_v
        self.pardict['c2'] *= uv_slope
        self.selfupdate(var=var)


class FM07(GenericCurve):
    """Extinction curves from Fitzpatrick and Massa 2007
    """

    def __init__(self, var=False):
        self.default_pars(var=var)
        self._f = self.f_07

    def default_pars(self, var=False):
        """FM07 parameterization coefficients, based on MW stars.

        :param var:
            Set to True to include scatter in the coefficients
            based on the FM07 sample.
        """
        p = {}
        # UV
        p['c2'] = 0.81
        p['c3'] = 2.99
        p['c4'] = 0.32
        p['c5'] = 6.10
        p['x0'] = 4.59
        p['gamma'] = 0.90
        # Optical
        p['ospline_x'] = 1e4/np.array([3300., 4000., 5330])
        p['ospline_k'] = (np.array([2.05, 1.32, 0.0]))
        # NIR
        p['alpha'] = 0-1.84
        p['R_v'] = 3.1

        if var:
            draws = normal(0, 1, 10)
            p['c2'] += draws[0] * 0.25
            p['c3'] += draws[1] * 1.0
            p['c4'] += draws[2] * 0.16
            p['c5'] += draws[3] * 0.6
            p['x0'] += draws[4] * 0.025
            p['gamma'] += draws[5] * 0.1
            p['ospline_k'] += draws[6:9] * np.array([0.17, 0.026, 0.008])
            p['R_v'] += draws[9] * 0.6

        self.pardict = p
        self.selfupdate(var=var)

    def selfupdate(self, var=False):
        """Enforce empirical relationships between some of the
        exctinction curve parameters, optionally including variance in
        the relationships.  These are based on data in FM09 and Gordon
        et al 2002.
        """
        k = -0.83 + 0.63 * self.pardict['R_v'] + var * normal(0, 0.12)
        c1 = 2.02 - 3.007 * self.pardict['c2'] + var * normal(0, 0.3)
        self.pardict['k'] = k
        self.pardict['c1'] = c1

    def ecurve(self, x, pardict):
        """Excess curve, including 3 components (UV, Optical, and
        NIR).
        """
        nir = self.powerlaw(x, **pardict)
        uv = self.fm_curve(x, **pardict)
        spline_x = np.array([1e4/2600, 1e4/2700] +
                            pardict['ospline_x'].tolist() +
                            [1.0, 0.75])
        spline_k = np.array(self.fm_curve(spline_x[0:2], **pardict).tolist() +
                            pardict['ospline_k'].tolist() +
                            self.powerlaw(spline_x[-2:], **pardict).tolist())

        optical = self.spline(x, spline_x=spline_x[::-1],
                              spline_k=spline_k[::-1])
        return (nir * (x < 0.75) + uv * (x > 3.8) +
                optical * ((x >= 0.75) & (x <= 3.8)))


class LMC(FM07):

    def __init__(self, var=False):
        self.default_pars(var=var)
        self._f = self.f_90

    def default_pars(self, var=False):
        """Gordon 2003 parameterization coefficients.  Set var to True
        to include scatter in the coefficients based on the Gordon 03
        sample.
        """
        p = {}
        # UV
        p['c2'] = 0.998 + var * normal(0, 0.027)
        p['c3'] = 2.719 + var * normal(0, 0.137)
        p['c4'] = 0.40 + var * normal(0, 0.036)
        p['x0'] = 4.579 + var * normal(0, 0.007)
        p['gamma'] = 0.934 + var * normal(0, 0.016)
        # Optical
        p['ospline_x'] = 1e4 / np.array([3300., 4000., 5330])
        p['ospline_k'] = (np.array([2.05, 1.32, 0.0]) +
                          var * normal(0, 1, 3) *
                          np.array([0.17, 0.026, 0.008]))
        # NIR
        p['alpha'] = -1.84
        p['R_v'] = 3.1 + var * normal(0, 0.6)

        self.pardict = p
        self.selfupdate()


class SMC(FM07):

    def __init__(self, var=False):
        self.default_pars(var=var)
        self._f = self.f_90

    def default_pars(self, var=False):
        """Gordon 2003 parameterization coefficients.  Set var to True
        to include scatter in the coefficients based on the Gordon 03
        sample.
        """
        p = {}
        # UV
        p['c2'] = 2.264 + var * normal(0, 0.040)
        p['c3'] = 0.389 + var * normal(0, 0.110)
        p['c4'] = 0.46 + var * normal(0, 0.079)
        p['x0'] = 4.60
        p['gamma'] = 1.0
        # Optical
        p['ospline_x'] = 1e4 / np.array([3300., 4000., 5330])
        p['ospline_k'] = (np.array([2.05, 1.32, 0.0]) +
                          var * normal(0, 1, 3) *
                          np.array([0.17, 0.026, 0.008]))
        # NIR
        p['alpha'] = -1.84
        p['R_v'] = 3.1 + var * normal(0, 0.6)

        self.pardict = p
        self.selfupdate()


class F99(GenericCurve):
    """Fitzpatrick 1999 R_v dependent extinction curves.  These are a
    one parameter family, though in practice we allow the bump
    strength to vary.
    """

    def __init__(self, var=False):
        self.default_pars(var=var)
        self._f = self.f_90

    def default_pars(self, var=False):
        """FM99 parameterization coefficients, for an R_v dependent
        curve.  Set var to True to include scatter in the coefficients
        based on the FM07 sample.
        """
        p = {}
        # NIR
        p['R_v'] = 3.1 + var * normal(0, 0.6)
        # Optical
        p['ospline_x'] = 1e4 / np.array([4100., 4670., 5470., 6000.,
                                         12200, 26500, 3e8])
        p['ospline_x'][-1] = 0.

        # UV
        p['c3'] = 3.23 + var * normal(0, 1.0)
        p['c4'] = 0.41 + var * normal(0, 0.16)
        p['x0'] = 4.596 + var * normal(0, 0.025)
        p['gamma'] = 0.99 + var * normal(0, 0.1)

        self.pardict = p
        self.selfupdate()

    def selfupdate(self, var=False):
        """Enforce empirical relationships between some of the
        exctinction curve parameters.
        """
        p = {}
        p['R_v'] = self.pardict['R_v']
        self.pardict['c2'] = (-0.824 + 4.717 / p['R_v'] +
                              var*normal(0, 0.25))
        self.pardict['c1'] = (2.02 - 3.007 * self.pardict['c2'] +
                              var * normal(0, 0.3))
        sk = np.array([1.208 + 0.0032 * p['R_v'] - 0.00033 * p['R_v']**2,
                       0.701 + 0.0016 * p['R_v'],
                       -0.050 + 0.0016 * p['R_v'],
                       -0.426 + 0.0044 * p['R_v'],
                       (0.829 / 3.1 - 1) * p['R_v'],
                       (0.265 / 3.1 - 1) * p['R_v'],
                       -p['R_v']])
        self.pardict['ospline_k'] = sk

    def ecurve(self, x, pardict):
        """Excess curve, including 2 components (UV and Optical/NIR)
        """
        uv = self.fm_curve(x, **pardict) * (x > 3.8)
        spline_x = np.array([1e4 / 2600, 1e4 / 2700] +
                            pardict['ospline_x'].tolist())
        spline_k = np.array(self.fm_curve(spline_x[0:2], **pardict).tolist() +
                            pardict['ospline_k'].tolist())
        optical = self.spline(x, spline_x=spline_x[::-1],
                              spline_k=spline_k[::-1]) * (x <= 3.8)
        return uv + optical
